import functools
from typing import Optional

import torch
from pytorch_lightning.cli import LightningArgumentParser, LRSchedulerCallable, OptimizerCallable
from pytorch_lightning.strategies import DDPFullyShardedNativeStrategy, StrategyRegistry
from torch.distributed.fsdp.wrap import transformer_auto_wrap_policy

from perceiver.model.core import CrossAttentionLayer, SelfAttentionLayer
from perceiver.model.text.clm import CausalLanguageModel, LitCausalLanguageModel
from perceiver.scripts.cli import CLI
from perceiver.scripts.lrs import *  # noqa: F403


# https://pytorch-lightning.readthedocs.io/en/1.9.0/advanced/model_parallel.html#fully-sharded-training
# (see NOTE box) is the reason for a custom configure_optimizers implementation below. clm_fsdp.py will
# be removed once this is fixed and clm.py will support FSDP directly.


torch.set_float32_matmul_precision("high")


policy = functools.partial(
    transformer_auto_wrap_policy,
    transformer_layer_cls={CrossAttentionLayer, SelfAttentionLayer, CausalLanguageModel},
)

StrategyRegistry.register(
    name="fsdp_perceiver_ar",
    strategy=DDPFullyShardedNativeStrategy,
    description="FSDP strategy optimized for Perceiver AR models",
    activation_checkpointing=[CrossAttentionLayer, SelfAttentionLayer],
    auto_wrap_policy=policy,
    cpu_offload=False,
)


class LitCausalLanguageModelFSDP(LitCausalLanguageModel):
    def __init__(
        self,
        optimizer: OptimizerCallable,
        scheduler: LRSchedulerCallable,
        *args,
        max_grad_norm: Optional[float] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.optimizer_fn = optimizer
        self.scheduler_fn = scheduler

    def configure_optimizers(self):
        optimizer = self.optimizer_fn(self.trainer.model.parameters())
        scheduler = self.scheduler_fn(optimizer)

        if isinstance(scheduler, CosineWithWarmupLR):  # noqa: F405
            scheduler.training_steps = self.trainer.estimated_stepping_batches

        return {
            "optimizer": optimizer,
            "lr_scheduler": {"scheduler": scheduler, "interval": "step", "frequency": 1},
        }

    def on_before_optimizer_step(self, optimizer, optimizer_idx):
        if self.hparams.max_grad_norm is not None:
            self.trainer.model.clip_grad_norm_(self.hparams.max_grad_norm)


class CausalLanguageModelCLI(CLI):
    def add_arguments_to_parser(self, parser: LightningArgumentParser) -> None:
        super().add_arguments_to_parser(parser)
        parser.link_arguments("data.max_seq_len", "model.max_seq_len", apply_on="instantiate")
        parser.link_arguments("data.vocab_size", "model.vocab_size", apply_on="instantiate")


if __name__ == "__main__":
    CausalLanguageModelCLI(
        LitCausalLanguageModelFSDP,
        auto_configure_optimizers=False,
        save_config_callback=None,
        run=True,
    )



import os
from typing import Any

from datasets import DatasetDict, load_dataset

from perceiver.data.text.common import Task, TextDataModule


class ImdbDataModule(TextDataModule):
    def __init__(
        self,
        *args: Any,
        dataset_dir: str = os.path.join(".cache", "imdb"),
        **kwargs: Any,
    ):
        super().__init__(dataset_dir, *args, **kwargs)

    @property
    def num_classes(self):
        return 2

    def load_source_dataset(self) -> DatasetDict:
        dataset = load_dataset("imdb", "plain_text", cache_dir=self.hparams.dataset_dir)

        if self.hparams.task == Task.clf:
            ds_train = dataset["train"]
            ds_valid = dataset["test"]
        else:
            column_to_remove = "label"
            ds_train = dataset["unsupervised"].remove_columns(column_to_remove)
            ds_valid = dataset["test"].remove_columns(column_to_remove)

        return DatasetDict(train=ds_train, valid=ds_valid)



import torch
import torch.nn as nn
from einops import rearrange

from perceiver.model.core.position import FrequencyPositionEncoding, positions


class InputAdapter(nn.Module):
    def __init__(self, num_input_channels: int, *args, **kwargs):
        """Transforms and position-encodes task-specific input to generic encoder input.

        :param num_input_channels: Number of channels of the generic encoder input produced by this adapter.
        """
        super().__init__()
        self._num_input_channels = num_input_channels

    @property
    def num_input_channels(self):
        return self._num_input_channels


class RotarySupport(InputAdapter):
    def __init__(self, rotated_channels_per_head: int, *args, **kwargs):
        """An input adapter mixin that additionally generates a frequency position encoding for input sequence
        `x`."""
        super().__init__(*args, **kwargs)
        self.frq_pos_encoding = FrequencyPositionEncoding(dim=rotated_channels_per_head)

    def forward(self, x, abs_pos=None):
        if abs_pos is None:
            abs_pos = positions(*x.shape, device=x.device)
        return super().forward(x, abs_pos), self.frq_pos_encoding(abs_pos)


class OutputAdapter(nn.Module):
    """Transforms generic decoder cross-attention output to task-specific output."""


class QueryProvider:
    """Provider of cross-attention query input."""

    @property
    def num_query_channels(self):
        raise NotImplementedError()

    def __call__(self, x=None):
        raise NotImplementedError()


class TrainableQueryProvider(nn.Module, QueryProvider):
    """Provider of learnable cross-attention query input.

    This is the latent array in Perceiver IO encoders and the output query array in most Perceiver IO decoders.
    """

    def __init__(self, num_queries: int, num_query_channels: int, init_scale: float = 0.02):
        super().__init__()
        self._query = nn.Parameter(torch.empty(num_queries, num_query_channels))
        self._init_parameters(init_scale)

    def _init_parameters(self, init_scale: float):
        with torch.no_grad():
            self._query.normal_(0.0, init_scale)

    @property
    def num_query_channels(self):
        return self._query.shape[-1]

    def forward(self, x=None):
        return rearrange(self._query, "... -> 1 ...")



import os

from dataclasses import dataclass
from typing import Any, Tuple

import torch
from einops import rearrange
from transformers import (
    PerceiverConfig as HuggingfacePerceiverConfig,
    PerceiverForImageClassificationFourier,
    PerceiverModel,
)

from perceiver.model.core import (
    EncoderConfig,
    FourierPositionEncoding,
    InputAdapter,
    PerceiverEncoder,
    PerceiverIO,
    PerceiverIOConfig,
    TrainableQueryProvider,
)
from perceiver.model.core.classifier import (
    ClassificationDecoder,
    ClassificationDecoderConfig,
    ClassificationOutputAdapter,
    LitClassifier,
)
from perceiver.model.core.convert import (
    copy_cross_attention_layer_params,
    copy_latent_provider_params,
    copy_self_attention_block_params,
)


@dataclass
class ImageEncoderConfig(EncoderConfig):
    image_shape: Tuple[int, int, int] = (224, 224, 3)
    num_frequency_bands: int = 32


class ImageInputAdapter(InputAdapter):
    def __init__(self, image_shape: Tuple[int, ...], num_frequency_bands: int):
        *spatial_shape, num_image_channels = image_shape
        position_encoding = FourierPositionEncoding(input_shape=spatial_shape, num_frequency_bands=num_frequency_bands)

        super().__init__(num_input_channels=num_image_channels + position_encoding.num_position_encoding_channels())

        self.image_shape = image_shape
        self.position_encoding = position_encoding

    def forward(self, x):
        b, *d = x.shape

        if tuple(d) != self.image_shape:
            raise ValueError(f"Input vision shape {tuple(d)} different from required shape {self.image_shape}")

        x_enc = self.position_encoding(b)
        x = rearrange(x, "b ... c -> b (...) c")
        return torch.cat([x, x_enc], dim=-1)


class ImageClassifier(PerceiverIO):
    def __init__(self, config: PerceiverIOConfig[ImageEncoderConfig, ClassificationDecoderConfig]):
        input_adapter = ImageInputAdapter(
            image_shape=config.encoder.image_shape,
            num_frequency_bands=config.encoder.num_frequency_bands,
        )

        encoder_kwargs = config.encoder.base_kwargs()
        if encoder_kwargs["num_cross_attention_qk_channels"] is None:
            encoder_kwargs["num_cross_attention_qk_channels"] = input_adapter.num_input_channels

        encoder = ImageEncoder(
            input_adapter=input_adapter,
            num_latents=config.num_latents,
            num_latent_channels=config.num_latent_channels,
            activation_checkpointing=config.activation_checkpointing,
            activation_offloading=config.activation_offloading,
            **encoder_kwargs,
        )
        output_query_provider = TrainableQueryProvider(
            num_queries=1,
            num_query_channels=config.decoder.num_output_query_channels,
            init_scale=config.decoder.init_scale,
        )
        output_adapter = ClassificationOutputAdapter(
            num_classes=config.decoder.num_classes,
            num_output_query_channels=config.decoder.num_output_query_channels,
        )
        decoder = ClassificationDecoder(
            output_adapter=output_adapter,
            output_query_provider=output_query_provider,
            num_latent_channels=config.num_latent_channels,
            activation_checkpointing=config.activation_checkpointing,
            activation_offloading=config.activation_offloading,
            **config.decoder.base_kwargs(),
        )
        super().__init__(encoder, decoder)

        if config.params is None:
            pass
        elif os.path.isfile(config.params):
            self.load_state_dict(torch.load(config.params))
        else:
            # import model params from Hugging Face Perceiver
            src = PerceiverForImageClassificationFourier.from_pretrained(config.params)
            self.encoder.copy_params(src.perceiver)
            self.decoder.copy_params(src.perceiver)


class ImageEncoder(PerceiverEncoder):
    def copy_params(self, src: PerceiverModel):
        copy_cross_attention_layer_params(src.encoder.cross_attention, self.cross_attn_1, query_residual=True)
        copy_self_attention_block_params(src.encoder.self_attends, self.self_attn_1)
        copy_latent_provider_params(src, self)


class LitImageClassifier(LitClassifier):
    def __init__(self, encoder: ImageEncoderConfig, decoder: ClassificationDecoderConfig, *args: Any, **kwargs: Any):
        super().__init__(encoder, decoder, *args, **kwargs)
        self.model = ImageClassifier(
            PerceiverIOConfig(
                encoder=encoder,
                decoder=decoder,
                num_latents=self.hparams.num_latents,
                num_latent_channels=self.hparams.num_latent_channels,
                activation_checkpointing=self.hparams.activation_checkpointing,
                activation_offloading=self.hparams.activation_offloading,
                params=self.hparams.params,
            )
        )

    def step(self, batch):
        x, y = batch["image"], batch["label"]
        return self.loss_acc(self(x), y)

    def forward(self, x):
        return self.model(x)


def convert_config(
    config: HuggingfacePerceiverConfig,
) -> PerceiverIOConfig[ImageEncoderConfig, ClassificationDecoderConfig]:
    assert config.hidden_act == "gelu"

    encoder_config = ImageEncoderConfig(
        image_shape=(224, 224, 3),
        num_frequency_bands=64,
        num_cross_attention_heads=config.num_cross_attention_heads,
        num_self_attention_heads=config.num_self_attention_heads,
        num_self_attention_layers_per_block=config.num_self_attends_per_block,
        num_self_attention_blocks=config.num_blocks,
        dropout=config.attention_probs_dropout_prob,
        init_scale=config.initializer_range,
    )
    decoder_config = ClassificationDecoderConfig(
        num_classes=config.num_labels,
        num_output_query_channels=config.d_latents,
        num_cross_attention_heads=config.num_cross_attention_heads,
        cross_attention_residual=True,
        dropout=config.attention_probs_dropout_prob,
        init_scale=config.initializer_range,
    )
    return PerceiverIOConfig(
        encoder_config,
        decoder_config,
        num_latents=config.num_latents,
        num_latent_channels=config.d_latents,
        params=config.name_or_path,
    )



from pytorch_lightning.cli import LightningArgumentParser

from perceiver.model.text.clm import LitCausalLanguageModel
from perceiver.scripts.cli import CLI
from perceiver.scripts.lrs import CosineWithWarmupLR


class CausalLanguageModelCLI(CLI):
    def add_arguments_to_parser(self, parser: LightningArgumentParser) -> None:
        super().add_arguments_to_parser(parser)
        parser.add_lr_scheduler_args(CosineWithWarmupLR)
        parser.link_arguments("trainer.max_steps", "lr_scheduler.training_steps", apply_on="parse")
        parser.link_arguments("data.max_seq_len", "model.max_seq_len", apply_on="instantiate")
        parser.link_arguments("data.vocab_size", "model.vocab_size", apply_on="instantiate")
        parser.set_defaults(
            {
                "model.max_latents": 512,
                "model.num_channels": 512,
                "model.num_self_attention_layers": 8,
                "model.cross_attention_dropout": 0.5,
                "model.post_attention_dropout": 0.0,
            }
        )


if __name__ == "__main__":
    CausalLanguageModelCLI(LitCausalLanguageModel, run=True)



from builtins import NotImplementedError
from dataclasses import dataclass
from typing import Any

import torch.nn as nn
import torchmetrics as tm
import transformers

from perceiver.model.core.adapter import OutputAdapter
from perceiver.model.core.config import DecoderConfig
from perceiver.model.core.convert import copy_cross_attention_layer_params, copy_param, copy_params
from perceiver.model.core.lightning import LitPerceiverIO
from perceiver.model.core.modules import PerceiverDecoder


@dataclass
class ClassificationDecoderConfig(DecoderConfig):
    num_output_queries: int = 1
    num_output_query_channels: int = 256
    num_classes: int = 100


class ClassificationOutputAdapter(OutputAdapter):
    def __init__(
        self,
        num_classes: int,
        num_output_query_channels: int,
    ):
        super().__init__()
        self.linear = nn.Linear(num_output_query_channels, num_classes)

    def forward(self, x):
        return self.linear(x).squeeze(dim=1)


class ClassificationDecoder(PerceiverDecoder):
    def copy_params(self, src: transformers.PerceiverModel):
        copy_cross_attention_layer_params(
            src.decoder.decoder.decoding_cross_attention, self.cross_attn, query_residual=True
        )
        # Copy output adapter parameters
        copy_params(src.decoder.decoder.final_layer, self.output_adapter.linear)
        # Copy output query provider parameters
        copy_param(src.decoder.decoder.output_position_encodings.position_embeddings, self.output_query_provider._query)


class LitClassifier(LitPerceiverIO):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.loss = nn.CrossEntropyLoss()
        self.acc = tm.classification.accuracy.Accuracy()

    def step(self, batch):
        raise NotImplementedError()

    def loss_acc(self, logits, y):
        loss = self.loss(logits, y)
        y_pred = logits.argmax(dim=-1)
        acc = self.acc(y_pred, y)
        return loss, acc

    def training_step(self, batch, batch_idx):
        loss, acc = self.step(batch)
        self.log("train_loss", loss)
        self.log("train_acc", acc, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        loss, acc = self.step(batch)
        self.log("val_loss", loss, prog_bar=True, sync_dist=True)
        self.log("val_acc", acc, prog_bar=True, sync_dist=True)

    def test_step(self, batch, batch_idx):
        loss, acc = self.step(batch)
        self.log("test_loss", loss, sync_dist=True)
        self.log("test_acc", acc, sync_dist=True)



import os
from typing import Any

from datasets import DatasetDict, load_dataset

from perceiver.data.text.common import TextDataModule


class WikiTextDataModule(TextDataModule):
    def __init__(
        self,
        *args: Any,
        dataset_dir: str = os.path.join(".cache", "wikitext"),
        **kwargs: Any,
    ):
        super().__init__(dataset_dir, *args, **kwargs)

    def load_source_dataset(self) -> DatasetDict:
        dataset = load_dataset("wikitext", "wikitext-103-raw-v1", cache_dir=self.hparams.dataset_dir)
        return DatasetDict(train=dataset["train"], valid=dataset["validation"])



from perceiver.data.text import (
    BookCorpusDataModule,
    BookCorpusOpenDataModule,
    C4DataModule,
    Enwik8DataModule,
    ImdbDataModule,
    WikipediaDataModule,
    WikiTextDataModule,
)



import os
from typing import Any, Union

from datasets import DatasetDict, load_dataset

from perceiver.data.text.common import TextDataModule


class BookCorpusOpenDataModule(TextDataModule):
    def __init__(
        self,
        *args: Any,
        dataset_dir: str = os.path.join(".cache", "bookcorpusopen"),
        source_train_size: Union[float, int, None] = None,
        source_valid_size: Union[float, int, None] = 0.02,
        preproc_batch_size: int = 10,
        **kwargs: Any,
    ):
        super().__init__(dataset_dir, *args, preproc_batch_size=preproc_batch_size, **kwargs)

    def load_source_dataset(self) -> DatasetDict:
        dataset = load_dataset("bookcorpusopen", "plain_text", split="train", cache_dir=self.hparams.dataset_dir)
        dataset = self._train_valid_split(dataset, self.hparams.source_train_size, self.hparams.source_valid_size)

        columns_to_remove = ["title"]

        ds_train = dataset["train"].remove_columns(columns_to_remove)
        ds_valid = dataset["valid"].remove_columns(columns_to_remove)

        return DatasetDict(train=ds_train, valid=ds_valid)



from collections import defaultdict
from typing import Optional

import numpy as np
import torch

from transformers import (
    DataCollatorForLanguageModeling,
    DataCollatorWithPadding,
    DefaultDataCollator,
    PreTrainedTokenizerFast,
)
from transformers.utils import PaddingStrategy


class Collator:
    def collate(self, examples):
        raise NotImplementedError()

    def __call__(self, examples):
        result = self.collate(examples)
        return result["labels"], result["input_ids"], ~result["attention_mask"].type(torch.bool)


class DefaultCollator(Collator):
    label_keys = ["label", "labels"]

    def __init__(self, tokenizer: PreTrainedTokenizerFast, max_seq_len: Optional[int] = None):
        self.collator = DefaultDataCollator()
        self.tokenizer = tokenizer
        self.max_seq_len = max_seq_len

    def collate(self, examples):
        cur_length = max(len(example["input_ids"]) for example in examples)
        max_length = min(cur_length, self.max_seq_len)
        return self.collator([self._prepare(example, max_length=max_length) for example in examples])

    def _prepare(self, example, max_length):
        # FIXME: ensure proper handling of special tokens in example
        # Sequences longer than max_length are truncated including any
        # special tokens at the end of the sequence. These special tokens
        # must be preserved though. Setting add_special_tokens=true doesn't
        # work either because this would duplicate (some) special tokens
        # already contained in the input sequence.
        prepared = self._prepare_sequence(example["input_ids"], max_length)

        if "label_ids" in example:
            prepared_label_ids = self._prepare_sequence(example["label_ids"], max_length)
            prepared["label_ids"] = prepared_label_ids["input_ids"]

        for label_key in self.label_keys:
            if label_key in example:
                prepared[label_key] = example[label_key]

        return prepared

    def _prepare_sequence(self, sequence, max_length):
        return self.tokenizer.prepare_for_model(
            sequence,
            add_special_tokens=False,
            return_token_type_ids=False,
            padding=False if self.tokenizer.pad_token is None else PaddingStrategy.MAX_LENGTH,
            max_length=max_length,
            truncation=True,
        )


class WordMaskingCollator(Collator):
    def __init__(self, tokenizer: PreTrainedTokenizerFast, mask_prob: float = 0.15):
        self.collator = DataCollatorWithPadding(tokenizer)
        self.mask_token_id = tokenizer.mask_token_id
        self.vocab_size = tokenizer.vocab_size
        self.mask_prob = mask_prob

    def collate(self, examples):
        return self.collator(self.mask_words(examples))

    def mask_words(self, examples):
        """A modified version of whole word masking as described in https://huggingface.co/course/chapter7/3.

        The implementation in the linked document replaces words, randomly selected with `wwm_probability`, with mask
        tokens (one or more per word). The implementation here, however, only replaces 80% of selected words with mask
        tokens and replaces 10% with random words and leaves 10% unchanged.
        """
        for example in examples:
            self.mask_words_1(example)
        return examples

    def mask_words_1(self, example):
        # ------------------
        #  Mutates argument
        # ------------------

        word_ids = example.pop("word_ids")
        input_ids = example["input_ids"]
        labels = [-100] * len(input_ids)

        mapping = defaultdict(list)
        current_word_index = -1
        current_word_id = None
        for idx, word_id in enumerate(word_ids):
            if word_id is not None:
                if word_id != current_word_id:
                    current_word_id = word_id
                    current_word_index += 1
                mapping[current_word_index].append(idx)

        # Randomly mask words
        mask = np.random.binomial(1, self.mask_prob, len(mapping))
        for word_index in np.where(mask)[0]:
            rand_nr = np.random.rand(2)
            for idx in mapping[word_index]:
                labels[idx] = input_ids[idx]
                if rand_nr[0] < 0.8:
                    # in 80% of cases replace word with mask token(s)
                    input_ids[idx] = self.mask_token_id
                elif rand_nr[1] < 0.5:
                    # in 10% of cases replace word token(s) with random tokens
                    input_ids[idx] = np.random.randint(self.vocab_size)
                else:
                    # in 10% of cases leave word token(s) unchanged
                    pass

        example["labels"] = labels
        return example


class TokenMaskingCollator(Collator):
    def __init__(self, tokenizer: PreTrainedTokenizerFast, mask_prob=0.15):
        self.collator = DataCollatorForLanguageModeling(tokenizer, mlm_probability=mask_prob)

    def collate(self, examples):
        return self.collator(examples)



from dataclasses import asdict, dataclass, fields
from typing import Generic, Optional, TypeVar


@dataclass
class EncoderConfig:
    num_cross_attention_heads: int = 8
    num_cross_attention_qk_channels: Optional[int] = None
    num_cross_attention_v_channels: Optional[int] = None
    num_cross_attention_layers: int = 1
    first_cross_attention_layer_shared: bool = False
    cross_attention_widening_factor: int = 1
    num_self_attention_heads: int = 8
    num_self_attention_qk_channels: Optional[int] = None
    num_self_attention_v_channels: Optional[int] = None
    num_self_attention_layers_per_block: int = 8
    num_self_attention_blocks: int = 1
    first_self_attention_block_shared: bool = True
    self_attention_widening_factor: int = 1
    dropout: float = 0.0
    init_scale: float = 0.02
    freeze: bool = False

    def base_kwargs(self, exclude=("freeze",)):
        return _base_kwargs(self, EncoderConfig, exclude)


@dataclass
class DecoderConfig:
    num_cross_attention_heads: int = 8
    num_cross_attention_qk_channels: Optional[int] = None
    num_cross_attention_v_channels: Optional[int] = None
    cross_attention_widening_factor: int = 1
    cross_attention_residual: bool = True
    dropout: float = 0.0
    init_scale: float = 0.02
    freeze: bool = False

    def base_kwargs(self, exclude=("freeze",)):
        return _base_kwargs(self, DecoderConfig, exclude)


E = TypeVar("E", bound=EncoderConfig)
D = TypeVar("D", bound=DecoderConfig)


@dataclass
class PerceiverIOConfig(Generic[E, D]):
    encoder: E
    decoder: D
    num_latents: int
    num_latent_channels: int
    activation_checkpointing: bool = False
    activation_offloading: bool = False
    params: Optional[str] = None


@dataclass
class PerceiverARConfig:
    num_heads: int = 8
    max_heads_parallel: Optional[int] = None
    num_self_attention_layers: int = 8
    self_attention_widening_factor: int = 4
    cross_attention_widening_factor: int = 4
    cross_attention_dropout: float = 0.5
    post_attention_dropout: float = 0.0
    activation_checkpointing: bool = False
    activation_offloading: bool = False

    def base_kwargs(self, exclude=()):
        return _base_kwargs(self, PerceiverARConfig, exclude)


def _base_kwargs(config, base_class, exclude):
    base_field_names = [field.name for field in fields(base_class) if field.name not in exclude]
    return {k: v for k, v in asdict(config).items() if k in base_field_names}



import os
from dataclasses import dataclass
from typing import Tuple

import torch
import torch.nn as nn
import transformers
from einops import rearrange

from perceiver.model.core import (
    EncoderConfig,
    FourierPositionEncoding,
    InputAdapter,
    OutputAdapter,
    PerceiverDecoder,
    PerceiverEncoder,
    PerceiverIO,
    PerceiverIOConfig,
    QueryProvider,
)
from perceiver.model.core.config import DecoderConfig
from perceiver.model.core.convert import (
    copy_cross_attention_layer_params,
    copy_latent_provider_params,
    copy_params,
    copy_self_attention_block_params,
)


@dataclass
class OpticalFlowEncoderConfig(EncoderConfig):
    image_shape: Tuple[int, int] = (368, 496)
    num_patch_input_channels: int = 27
    num_patch_hidden_channels: int = 64
    num_frequency_bands: int = 64


@dataclass
class OpticalFlowDecoderConfig(DecoderConfig):
    image_shape: Tuple[int, int] = (368, 496)
    rescale_factor: float = 100.0


class OpticalFlowInputAdapter(InputAdapter):
    def __init__(
        self,
        image_shape: Tuple[int, int],
        num_patch_input_channels: int,
        num_patch_hidden_channels: int,
        num_frequency_bands: int,
    ):
        position_encoding = FourierPositionEncoding(input_shape=image_shape, num_frequency_bands=num_frequency_bands)
        super().__init__(num_patch_hidden_channels + position_encoding.num_position_encoding_channels())

        self.linear = nn.Linear((num_patch_input_channels * 2), num_patch_hidden_channels)
        self.position_encoding = position_encoding

    def forward(self, x):
        b, *_ = x.shape

        x = rearrange(x, "b t c h w -> b h w (t c)")  # concatenate temporal inputs in the channel dimension
        x = self.linear(x)
        x = rearrange(x, "b ... c -> b (...) c")
        pos_enc = self.position_encoding(b)
        return torch.cat([x, pos_enc], dim=-1)


class OpticalFlowOutputAdapter(OutputAdapter):
    def __init__(
        self,
        image_shape: Tuple[int, int],
        num_output_query_channels: int,
        num_output_image_channels: int = 2,
        rescale_factor: float = 100.0,
    ):
        super().__init__()
        self.image_shape = image_shape
        self.rescale_factor = rescale_factor
        self.linear = nn.Linear(num_output_query_channels, num_output_image_channels)

    def forward(self, x):
        x = self.linear(x) / self.rescale_factor
        return rearrange(x, "b (h w) c -> b h w c", h=self.image_shape[0])


class OpticalFlowQueryProvider(nn.Module, QueryProvider):
    def __init__(self, num_query_channels: int):
        super().__init__()
        self._num_query_channels = num_query_channels

    @property
    def num_query_channels(self):
        return self._num_query_channels

    def forward(self, x):
        assert x.shape[-1] == self.num_query_channels
        return x


class OpticalFlow(PerceiverIO):
    def __init__(self, config: PerceiverIOConfig[OpticalFlowEncoderConfig, OpticalFlowDecoderConfig]):
        input_adapter = OpticalFlowInputAdapter(
            image_shape=config.encoder.image_shape,
            num_patch_input_channels=config.encoder.num_patch_input_channels,
            num_patch_hidden_channels=config.encoder.num_patch_hidden_channels,
            num_frequency_bands=config.encoder.num_frequency_bands,
        )

        encoder_kwargs = config.encoder.base_kwargs()
        if encoder_kwargs["num_cross_attention_qk_channels"] is None:
            encoder_kwargs["num_cross_attention_qk_channels"] = input_adapter.num_input_channels

        if encoder_kwargs["num_cross_attention_v_channels"] is None:
            encoder_kwargs["num_cross_attention_v_channels"] = input_adapter.num_input_channels

        encoder = OpticalFlowEncoder(
            input_adapter=input_adapter,
            num_latents=config.num_latents,
            num_latent_channels=config.num_latent_channels,
            activation_checkpointing=config.activation_checkpointing,
            activation_offloading=config.activation_offloading,
            **encoder_kwargs,
        )
        output_adapter = OpticalFlowOutputAdapter(
            image_shape=config.decoder.image_shape,
            num_output_query_channels=input_adapter.num_input_channels,
            rescale_factor=config.decoder.rescale_factor,
        )
        output_query_provider = OpticalFlowQueryProvider(num_query_channels=input_adapter.num_input_channels)
        decoder = OpticalFlowDecoder(
            output_adapter=output_adapter,
            output_query_provider=output_query_provider,
            num_latent_channels=config.num_latent_channels,
            activation_checkpointing=config.activation_checkpointing,
            activation_offloading=config.activation_offloading,
            **config.decoder.base_kwargs(),
        )
        super().__init__(encoder, decoder)

        if config.params is None:
            pass
        elif os.path.isfile(config.params):
            self.load_state_dict(torch.load(config.params))
        else:
            # import model params from Huggingface Perceiver
            src = transformers.PerceiverForOpticalFlow.from_pretrained(config.params)
            self.encoder.copy_params(src.perceiver)
            self.decoder.copy_params(src.perceiver)

    def forward(self, x: torch.Tensor):
        x_latent, x_adapted = self.encoder(x, return_adapted_input=True)
        return self.decoder(x_latent, x_adapted=x_adapted)


class OpticalFlowEncoder(PerceiverEncoder):
    def copy_params(self, src: transformers.PerceiverModel):
        copy_cross_attention_layer_params(src.encoder.cross_attention, self.cross_attn_1, query_residual=True)
        copy_self_attention_block_params(src.encoder.self_attends, self.self_attn_1)
        copy_latent_provider_params(src, self)
        # Copy input adapter parameters
        copy_params(src.input_preprocessor.conv_after_patches, self.input_adapter.linear)


class OpticalFlowDecoder(PerceiverDecoder):
    def copy_params(self, src: transformers.PerceiverModel):
        copy_cross_attention_layer_params(
            src.decoder.decoder.decoding_cross_attention, self.cross_attn, query_residual=False
        )
        # Copy output adapter parameters
        copy_params(src.decoder.decoder.final_layer, self.output_adapter.linear)


def convert_config(
    config: transformers.PerceiverConfig,
) -> PerceiverIOConfig[OpticalFlowEncoderConfig, OpticalFlowDecoderConfig]:
    assert config.hidden_act == "gelu"

    image_shape = tuple(config.train_size)

    encoder_config = OpticalFlowEncoderConfig(
        image_shape=image_shape,
        num_patch_input_channels=27,
        num_patch_hidden_channels=64,
        num_frequency_bands=64,
        num_cross_attention_layers=1,
        num_cross_attention_heads=config.num_cross_attention_heads,
        num_self_attention_heads=config.num_self_attention_heads,
        num_self_attention_layers_per_block=config.num_self_attends_per_block,
        num_self_attention_blocks=config.num_blocks,
        first_self_attention_block_shared=True,
        cross_attention_widening_factor=config.cross_attention_widening_factor,
        self_attention_widening_factor=config.self_attention_widening_factor,
        dropout=config.attention_probs_dropout_prob,
        init_scale=config.initializer_range,
    )
    decoder_config = OpticalFlowDecoderConfig(
        image_shape=image_shape,
        num_cross_attention_qk_channels=512,
        num_cross_attention_v_channels=512,
        num_cross_attention_heads=config.num_cross_attention_heads,
        cross_attention_widening_factor=config.cross_attention_widening_factor,
        cross_attention_residual=False,
        dropout=config.attention_probs_dropout_prob,
        init_scale=config.initializer_range,
        rescale_factor=100.0,
    )
    return PerceiverIOConfig(
        encoder_config,
        decoder_config,
        num_latents=config.num_latents,
        num_latent_channels=config.d_latents,
        params=config.name_or_path,
    )



from dataclasses import asdict, dataclass, fields
from typing import Any, Optional

import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.utilities import rank_zero_only
from tqdm import tqdm

from perceiver.model.core import OutputAdapter, PerceiverAR, PerceiverARConfig, RotarySupport
from perceiver.model.core.utils import init_parameters
from perceiver.model.text import common


@dataclass
class CausalLanguageModelConfig(PerceiverARConfig):
    vocab_size: int = 262
    max_seq_len: int = 4096
    max_latents: int = 512
    num_channels: int = 512
    output_norm: bool = False
    init_scale: float = 0.02

    @classmethod
    def create(cls, **kwargs):
        return cls(**{field.name: kwargs[field.name] for field in fields(cls) if field.name in kwargs})


class TextInputAdapter(RotarySupport, common.TextInputAdapter):
    def __init__(self, rotated_channels_per_head: int, vocab_size: int, max_seq_len: int, num_input_channels: int):
        super().__init__(rotated_channels_per_head, vocab_size, max_seq_len, num_input_channels)

    def forward(self, x, abs_pos=None):
        return super().forward(x, abs_pos)


class TextOutputAdapter(OutputAdapter):
    def __init__(self, num_channels: int, vocab_size: int):
        super().__init__()
        self.linear = nn.Linear(num_channels, vocab_size)

    def forward(self, x):
        return self.linear(x)


class CausalLanguageModel(PerceiverAR):
    def __init__(self, config: CausalLanguageModelConfig):
        input_adapter = TextInputAdapter(
            # Rotary position embedding for first 50% of channels ...
            rotated_channels_per_head=config.num_channels // config.num_heads // 2,
            vocab_size=config.vocab_size,
            max_seq_len=config.max_seq_len,
            num_input_channels=config.num_channels,
        )
        super().__init__(input_adapter=input_adapter, **config.base_kwargs())
        self._config = config

        if config.output_norm:
            self.out_norm = nn.LayerNorm(config.num_channels)

        self.output_adapter = common.TiedTextOutputAdapter(vocab_size=config.vocab_size)
        self._init_parameters(config.init_scale)

    def _init_parameters(self, init_scale: float):
        with torch.no_grad():
            init_parameters(self, init_scale)

    @property
    def max_seq_len(self):
        return self.input_adapter.max_seq_len

    @property
    def max_latents(self):
        return self._config.max_latents

    @property
    def max_prefix_len(self):
        return self.max_seq_len - self.max_latents

    def forward(self, x, prefix_len, pad_mask=None):
        if prefix_len > self.max_prefix_len:
            raise ValueError(f"prefix_len ({prefix_len}) exceeds max_prefix_len ({self.max_prefix_len})")

        x_latent = super().forward(x, prefix_len, pad_mask)

        if self._config.output_norm:
            x_latent = self.out_norm(x_latent)

        return self.output_adapter(x_latent, txt_embedding=self.input_adapter.txt_embedding)

    @torch.no_grad()
    def generate(
        self,
        prompt: torch.Tensor,
        pad_mask: Optional[torch.Tensor] = None,
        num_tokens: int = 512,
        num_latents: int = 1,
        top_k: int = 5,
        temperature: float = 1.0,
        pbar: bool = True,
    ):
        """Generate sequence from `prompt` via `top-k` sampling at given `temperature`.

        :param prompt: Prompt of shape (B, N). If sequences have different length they must be left-padded.
        :param pad_mask: Prompt pad mask of shape (B, N). Must be supplied if prompt contains pad tokens.
        :param num_tokens: Number of tokens to generate.
        :param num_latents: Initial number of latent positions.
        :param top_k: Number of most likely tokens to sample from.
        :param temperature: "Controls the entropy of next token probabilities."
        :param pbar: If `True`, uses a progress bar during generation.
        """

        n_init = prompt.shape[1]

        if 0 <= num_latents <= self.max_latents:
            prefix_len = n_init - num_latents
        else:
            raise ValueError(f"num_latents ({num_latents}) must be in range [0..{self.max_latents}]")

        result = prompt
        result_pad_mask = pad_mask

        num_tokens_range = range(num_tokens)

        if pbar:
            num_tokens_range = tqdm(num_tokens_range)

        for _ in num_tokens_range:
            if result.shape[1] - prefix_len == self.max_latents and prefix_len < self.max_prefix_len:
                # num_latents == max_latents reached, but not max_prefix_len yet.
                # Extend prefix by 1 token and keep num_latents == max_latents.
                prefix_len += 1

            logits = self(result[:, -self.max_seq_len :], prefix_len=prefix_len, pad_mask=result_pad_mask)[:, -1]
            logits = self.top_k(logits, top_k)

            probs = F.softmax(logits / temperature, dim=-1)
            sample = torch.multinomial(probs, 1)
            result = torch.cat((result, sample), dim=-1)

            if result_pad_mask is not None:
                sample_pad_mask = torch.zeros_like(sample, dtype=torch.bool)
                result_pad_mask = torch.cat((result_pad_mask, sample_pad_mask), dim=-1)
                result_pad_mask = result_pad_mask[:, -self.max_seq_len :]

        return result[:, n_init:]

    @staticmethod
    def top_k(logits: torch.Tensor, k: int):
        val, idx = torch.topk(logits, k)
        logits_top = torch.full_like(logits, float("-inf"))
        logits_top.scatter_(1, idx, val)
        return logits_top


class LitCausalLanguageModel(pl.LightningModule):
    def __init__(
        self,
        vocab_size: int,
        max_seq_len: int,
        max_latents: int = 512,
        num_channels: int = 512,
        num_heads: int = 8,
        max_heads_parallel: Optional[int] = None,
        num_self_attention_layers: int = 6,
        self_attention_widening_factor: int = 4,
        cross_attention_widening_factor: int = 4,
        cross_attention_dropout: float = 0.5,
        post_attention_dropout: float = 0.0,
        output_norm: bool = False,
        init_scale: float = 0.02,
        activation_checkpointing=False,
        activation_offloading=False,
        validation_sample_prompt: Optional[str] = None,
        validation_sample_record: Optional[int] = None,
    ):
        super().__init__()
        self.save_hyperparameters()
        self.model = CausalLanguageModel(CausalLanguageModelConfig.create(**self.hparams))
        self.loss = nn.CrossEntropyLoss()

    @classmethod
    def create(cls, config: CausalLanguageModelConfig, **kwargs: Any):
        return cls(**asdict(config), **kwargs)

    def setup(self, stage: Optional[str] = None):
        dm = self.trainer.datamodule

        if dm.tokenizer.pad_token is not None and dm.tokenizer.padding_side != "left":
            raise ValueError(
                "Causal language modeling with Perceiver AR requires a data module configured with padding_side=left"
            )

        self.preprocessor = dm.text_preprocessor()
        self.tokenizer = dm.tokenizer
        self.ds_valid = dm.ds_valid

    def forward(self, x, prefix_len, pad_mask=None):
        return self.model(x, prefix_len=prefix_len, pad_mask=pad_mask)

    def step(self, batch):
        labels, x, pad_mask = batch
        labels[pad_mask] = -100

        seq_len = x.shape[1]
        max_lat = self.hparams.max_latents

        if seq_len < max_lat:
            raise ValueError(f"Training sequence length must be at least {max_lat} (= nax_latents)")

        logits = self(x, prefix_len=seq_len - max_lat, pad_mask=pad_mask)
        labels = labels[:, -logits.shape[1] :]

        logits = rearrange(logits, "b n c -> (b n) c")
        labels = rearrange(labels, "b n -> (b n)")

        return self.loss(logits, labels)

    def training_step(self, batch, batch_idx):
        loss = self.step(batch)
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        loss = self.step(batch)
        self.log("val_loss", loss, prog_bar=True, sync_dist=True)

    @rank_zero_only
    def on_validation_epoch_end(self) -> None:
        if self.hparams.validation_sample_record is not None:
            if self.hparams.validation_sample_record == -1:
                # pick a random record from ds_valid as prompt
                record_idx = torch.randint(len(self.ds_valid), (1,)).item()
            else:
                # pick the specified record from ds_valid as prompt
                record_idx = self.hparams.validation_sample_record

            prompt = self.ds_valid[record_idx]["input_ids"]
            prompt_text = self.tokenizer.decode(prompt)
            prompt = torch.tensor(prompt).to(self.device)

            result = self.model.generate(
                num_tokens=512,
                prompt=prompt[None, ...],
                num_latents=self.hparams.max_latents,
                threshold=0.9,
                pbar=False,
            )
            result_text = self.tokenizer.decode(result[0])

            self.log_sample(tag="generated text (1)", prompt=prompt_text, generated=result_text)

        if self.hparams.validation_sample_prompt is not None:
            prompt_text = self.hparams.validation_sample_prompt
            prompt, _ = self.preprocessor.preprocess(prompt_text)
            prompt = prompt.to(self.device)

            result = self.model.generate(
                num_tokens=512, prompt=prompt[None, ...], num_latents=1, threshold=0.9, pbar=False
            )
            result_text = self.tokenizer.decode(result[0])

            self.log_sample(tag="generated text (2)", prompt=prompt_text, generated=result_text)

    def log_sample(self, tag, prompt, generated):
        if isinstance(self.logger, TensorBoardLogger):
            text = f"prompt:    {cleanup(prompt)}\n" f"generated: {cleanup(generated)}\n"
            self.logger.experiment.add_text(tag, f"<pre>{text}</pre>", self.trainer.global_step)
        else:
            # support other loggers here ...
            ...


def cleanup(text):
    return "".join([chr(max(32, ord(c))) for c in text])



import itertools
import math
from pathlib import Path
from typing import List, Tuple, Union

import cv2
import numpy as np
import torch
import torch.nn.functional as F
from einops import rearrange
from tqdm import tqdm

from perceiver.data.vision.video_utils import write_video


class OpticalFlowProcessor:
    def __init__(self, patch_size: Tuple[int, int], patch_min_overlap: int = 20, flow_scale_factor: int = 20):
        if patch_min_overlap >= patch_size[0] or patch_min_overlap >= patch_size[1]:
            raise ValueError(
                f"Overlap should be smaller than the patch size "
                f"(patch-size='{patch_size}', patch_min_overlap='{patch_min_overlap}')."
            )

        self.patch_size = patch_size
        self.patch_min_overlap = patch_min_overlap
        self.flow_scale_factor = flow_scale_factor

    @staticmethod
    def _to_tensor(x) -> torch.Tensor:
        if isinstance(x, torch.Tensor):
            return x
        if isinstance(x, np.ndarray):
            return torch.from_numpy(x)
        raise ValueError("Invalid input type. Provide input as np.array or torch.Tensor.")

    def _preprocess(
        self,
        image_pair: Union[Tuple[np.ndarray, np.ndarray], Tuple[torch.Tensor, torch.Tensor]],
        grid_indices: List[Tuple[int, int]],
    ) -> torch.Tensor:
        img1 = self._to_tensor(image_pair[0])
        img2 = self._to_tensor(image_pair[1])

        if img1.shape != img2.shape:
            raise ValueError(f"Shapes of images must match. (shape image1='{img1.shape}', shape image2='{img2.shape}')")

        height = img1.shape[0]
        width = img1.shape[1]
        if height < self.patch_size[0]:
            raise ValueError(
                f"Height of image (height='{height}') must be at least {self.patch_size[0]}."
                "Please pad or resize your image to the minimum dimension."
            )
        if width < self.patch_size[1]:
            raise ValueError(
                f"Width of image (width='{width}') must be at least {self.patch_size[1]}."
                "Please pad or resize your image to the minimum dimension."
            )

        img1 = self._transform(img1)
        img2 = self._transform(img2)
        transformed_image_pair = torch.stack([img1, img2], dim=0)

        patch_input_feature_batch = []
        for y, x in grid_indices:
            image_pair_patch = transformed_image_pair[..., y : y + self.patch_size[0], x : x + self.patch_size[1]]
            patch_input_features = self._extract_image_patches(image_pair_patch, kernel=3).float()
            patch_input_feature_batch.append(patch_input_features)

        return torch.stack(patch_input_feature_batch, dim=0)

    def _transform(self, img: torch.Tensor) -> torch.Tensor:
        x = self._normalize(img).to(torch.float32)

        if x.shape[-1] == 3:  # check if channels are at last position
            x = rearrange(x, "h w c -> c h w")
        return x

    @staticmethod
    def _normalize(img: torch.Tensor) -> torch.Tensor:
        return img / 255.0 * 2 - 1

    def _extract_image_patches(self, x: torch.Tensor, kernel: int, stride: int = 1, dilation: int = 1):
        """Equivalent to the implementation of https://www.tensorflow.org/api_docs/python/tf/image/extract_patches
        using "SAME" padding.

        From: https://discuss.pytorch.org/t/tf-extract-image-patches-in-pytorch/43837/9
        """
        b = x.shape[0]
        x = self._pad(x, kernel, stride, dilation)
        # extract patches
        patches = x.unfold(2, kernel, stride).unfold(3, kernel, stride)
        # re-order patch dimensions
        patches = patches.permute(0, 4, 5, 1, 2, 3).contiguous()
        # stack patches along second dimension
        return patches.view(b, -1, patches.shape[-2], patches.shape[-1])

    @staticmethod
    def _pad(x: torch.Tensor, kernel: int, stride: int = 1, dilation: int = 1) -> torch.Tensor:
        """Applies a pad to the input using "SAME" strategy."""
        *_, h, w = x.shape
        h2 = math.ceil(h / stride)
        w2 = math.ceil(w / stride)
        pad_row = (h2 - 1) * stride + (kernel - 1) * dilation + 1 - h
        pad_col = (w2 - 1) * stride + (kernel - 1) * dilation + 1 - w
        return F.pad(x, (pad_col // 2, pad_col - pad_col // 2, pad_row // 2, pad_row - pad_row // 2))

    def _compute_patch_grid_indices(self, img_shape: Tuple[int, ...]) -> List[Tuple[int, int]]:
        """From https://github.com/deepmind/deepmind-research/blob/master/perceiver/colabs/optical_flow.ipynb."""
        ys = list(range(0, img_shape[0], self.patch_size[0] - self.patch_min_overlap))
        xs = list(range(0, img_shape[1], self.patch_size[1] - self.patch_min_overlap))
        ys[-1] = img_shape[0] - self.patch_size[0]
        xs[-1] = img_shape[1] - self.patch_size[1]
        return list(itertools.product(ys, xs))

    def preprocess(
        self, image_pair: Union[Tuple[np.ndarray, np.ndarray], Tuple[torch.Tensor, torch.Tensor]]
    ) -> torch.Tensor:
        """Creates the input features for the model for a pair of images.

        The input images are stacked and split into image patches of size `patch_size`. For each pixel of each
        individual patch, 3x3 patches are extracted and stacked in the channel dimension.

        Output shape: torch.Size(nr_patches, 2, 27, patch_size[0], patch_size[1])
        """
        grid_indices = self._compute_patch_grid_indices(image_pair[0].shape)
        return self._preprocess(image_pair, grid_indices)

    def preprocess_batch(
        self,
        image_pairs: Union[List[Tuple[np.ndarray, np.ndarray]], List[Tuple[torch.Tensor, torch.Tensor]]],
    ) -> torch.Tensor:
        """Creates the input features for the model for a batch of image pairs.

        For each image pair the images are stacked and split into image patches of size `patch_size`. For each pixel
        of each individual patch, 3x3 patches are extracted and stacked in the channel dimension.

        Output shape: torch.Size(batch_size, nr_patches, 2, 27, patch_size[0], patch_size[1])
        """
        grid_indices = self._compute_patch_grid_indices(image_pairs[0][0].shape)
        return self._preprocess_batch(image_pairs, grid_indices)

    def _preprocess_batch(
        self,
        image_pairs: Union[List[Tuple[np.ndarray, np.ndarray]], List[Tuple[torch.Tensor, torch.Tensor]]],
        grid_indices: List[Tuple[int, int]],
    ) -> torch.Tensor:
        shapes = []
        for image1, image2 in image_pairs:
            shapes += [image1.shape, image2.shape]

        if not np.all(np.array(shapes) == shapes[0]):
            raise ValueError("Shapes of images must match. Not all input images have the same shape.")

        return torch.stack([self._preprocess(image_pair, grid_indices) for image_pair in image_pairs], dim=0)

    def postprocess(self, predictions: torch.Tensor, img_shape: Tuple[int, ...]) -> torch.Tensor:
        """Combines optical flow predictions for individual image patches into a single prediction per image pair.

        Predictions can be supplied for a single image pair or a batch of image pairs, hence the supported input shapes
        are:
        * (nr_patches, patch_size[0], patch_size[1], 2) and
        * (batch_size, nr_patches, patch_size[0], patch_size[1], 2).

        Returns combined predictions for each supplied image pair.

        Output shape: (batch_size, height, width, 2)
        """

        flow_batch = []
        height = img_shape[0]
        width = img_shape[1]

        grid_indices = self._compute_patch_grid_indices(img_shape)
        prediction_batch = predictions.unsqueeze(0).cpu() if predictions.dim() == 4 else predictions.cpu()

        b, p, *_ = prediction_batch.shape
        if p != len(grid_indices):
            raise ValueError(
                f"Number of patches in the input does not match the number of calculated patches based "
                f"on the supplied image size (nr_patches='{p}', calculated={len(grid_indices)})."
            )

        for prediction in prediction_batch:
            flow = torch.zeros(1, height, width, 2).type(torch.float32)
            flow_weights = torch.zeros(1, height, width, 1).type(torch.float32)

            for flow_patch, (y, x) in zip(prediction, grid_indices):
                flow_patch = flow_patch * self.flow_scale_factor

                weights_y, weights_x = torch.meshgrid(
                    torch.arange(self.patch_size[0]), torch.arange(self.patch_size[1]), indexing="ij"
                )
                weights_x = torch.minimum(torch.add(weights_x, 1), self.patch_size[1] - weights_x)
                weights_y = torch.minimum(torch.add(weights_y, 1), self.patch_size[0] - weights_y)
                weights = rearrange(torch.minimum(weights_x, weights_y), "h w -> 1 h w 1")

                pad = (0, 0, x, width - x - self.patch_size[1], y, height - y - self.patch_size[0], 0, 0)
                flow += F.pad(flow_patch * weights, pad, "constant", 0)
                flow_weights += F.pad(weights, pad, "constant", 0)

            flow /= flow_weights
            flow_batch.append(flow)

        return torch.concat(flow_batch, dim=0)

    def process(
        self,
        model,
        image_pairs: Union[List[Tuple[np.ndarray, np.ndarray]], List[Tuple[torch.Tensor, torch.Tensor]]],
        batch_size: int,
        device: torch.device,
    ) -> torch.Tensor:
        """Combines preprocessing, inference and postprocessing steps for the optical flow.

        The input features for model are created by stacking each image pair in the channel dimension and splitting the
        result into image patches of size `patch_size`. For each pixel in each individual patch, 3x3 patches
        are extracted and stacked in the channel dimension.

        The input is processed using the supplied optical flow model and the optical flow predictions per image pair
        are returned.

        Output shape: (batch_size, height, width, 2)
        """

        image_shape = image_pairs[0][0].shape
        grid_indices = self._compute_patch_grid_indices(image_shape)

        predictions = []
        with torch.no_grad():
            for i in tqdm(range(0, len(image_pairs), batch_size)):
                input_features_batch = self._preprocess_batch(image_pairs[i : i + batch_size], grid_indices)
                input_features_batch = rearrange(input_features_batch, "b p t c h w -> (b p) t c h w")
                for j in range(0, input_features_batch.shape[0], batch_size):
                    input_features_micro_batch = input_features_batch[j : (j + batch_size)]
                    input_features_micro_batch = input_features_micro_batch.to(device)
                    pred = model(input_features_micro_batch)
                    predictions.append(pred.cpu().detach())

        flow_predictions = torch.concat(predictions, dim=0)
        flow_predictions = rearrange(flow_predictions, "(b p) h w c -> b p h w c", b=len(image_pairs))
        return self.postprocess(flow_predictions, image_shape)


def render_optical_flow(flow: np.ndarray) -> np.ndarray:
    """Renders optical flow predictions produced by an optical flow model."""
    hsv = np.zeros((flow.shape[0], flow.shape[1], 3), dtype=np.uint8)

    mag, ang = cv2.cartToPolar(flow[..., 0], flow[..., 1])

    hsv[..., 0] = ang / np.pi / 2 * 180
    hsv[..., 1] = np.clip(mag * 255 / 24, 0, 255)
    hsv[..., 2] = 255

    return cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB)


def write_optical_flow_video(video_path: Path, frames: List[torch.Tensor], fps: int = 30) -> None:
    """Writes optical flow predictions as individual frames to a video file."""
    write_video(video_path=video_path, frames=[render_optical_flow(f.numpy()) for f in frames], fps=fps)



import os
from typing import Any, Union

from datasets import DatasetDict, load_dataset

from perceiver.data.text.common import TextDataModule


class Enwik8DataModule(TextDataModule):
    def __init__(
        self,
        *args: Any,
        dataset_dir: str = os.path.join(".cache", "enwik8"),
        source_train_size: Union[float, int, None] = None,
        source_valid_size: Union[float, int, None] = 0.05,
        **kwargs: Any,
    ):
        super().__init__(dataset_dir, *args, **kwargs)

    def load_source_dataset(self) -> DatasetDict:
        dataset = load_dataset("enwik8", "enwik8", split="train", cache_dir=self.hparams.dataset_dir)
        dataset = self._train_valid_split(dataset, self.hparams.source_train_size, self.hparams.source_valid_size)

        def append_newline(example):
            return {"text": example["text"] + "\n"}

        result = DatasetDict()
        for key in dataset.keys():
            result[key] = dataset[key].map(
                append_newline,
                num_proc=self.preproc_workers,
                desc="Append newline character",
            )
        return result



import os
from typing import Optional

import pytorch_lightning as pl
import torch
import torch.distributed as dist
from datasets import load_dataset
from datasets.distributed import split_dataset_by_node
from torch.utils.data import DataLoader
from transformers import AutoTokenizer

from perceiver.data.text.common import TextPreprocessor


os.environ["TOKENIZERS_PARALLELISM"] = "false"
os.environ["TRANSFORMERS_NO_ADVISORY_WARNINGS"] = "true"


class C4DataModule(pl.LightningDataModule):
    def __init__(
        self,
        tokenizer: str,
        max_seq_len: int,
        min_seq_len: Optional[int] = None,
        batch_size: int = 4,
        shuffle_window_seed: int = 0,
        shuffle_window_size: int = 10000,
        concat_batch_size: int = 16,
        num_train_workers: int = 2,
        num_valid_workers: int = 1,
        padding_side: Optional[str] = None,
        pin_memory: bool = True,
        rank: Optional[int] = None,
        world_size: Optional[int] = None,
    ):
        super().__init__()
        self.save_hyperparameters()
        self.tokenizer = AutoTokenizer.from_pretrained(self.hparams.tokenizer, verbose=False)
        self.collator = C4Collator(self.tokenizer)

        if self.hparams.padding_side is not None:
            self.tokenizer.padding_side = self.hparams.padding_side

        self.ds_train = None
        self.ds_valid = None

    @property
    def vocab_size(self):
        return self.tokenizer.vocab_size

    @property
    def max_seq_len(self):
        return self.hparams.max_seq_len

    @property
    def rank(self):
        # TODO: investigate if rank should be obtained from trainer
        return dist.get_rank() if self.hparams.rank is None else self.hparams.rank

    @property
    def world_size(self):
        # TODO: investigate if world_size should be obtained from trainer
        return dist.get_world_size() if self.hparams.world_size is None else self.hparams.world_size

    def current_epoch(self) -> int:
        return self.trainer.current_epoch if self.trainer else 0

    def text_preprocessor(self):
        return TextPreprocessor(
            tokenizer=self.hparams.tokenizer,
            max_seq_len=self.hparams.max_seq_len,
            add_special_tokens=False,
        )

    def _create_dataset(self, split):
        dataset = load_dataset("c4", "en", split=split, streaming=True)
        dataset = dataset.shuffle(seed=self.hparams.shuffle_window_seed, buffer_size=self.hparams.shuffle_window_size)
        return split_dataset_by_node(dataset, rank=self.rank, world_size=self.world_size)

    def _create_pipeline(self, dataset, min_seq_len=None):
        def tokenize(examples):
            return self.tokenizer(
                examples["text"],
                padding=False,
                truncation=False,
                max_length=None,
                add_special_tokens=False,
                return_token_type_ids=False,
                return_attention_mask=False,
            )

        def concat(examples):
            for example in examples:
                yield from example
                yield self.tokenizer.eos_token_id

        def chunk_len():
            if min_seq_len is None:
                return self.hparams.max_seq_len + 1
            else:
                return torch.randint(min_seq_len, self.hparams.max_seq_len + 1, size=(1,)) + 1

        def chunk(examples):
            chs = []
            ch = []
            ch_len = chunk_len()
            for token_id in concat(examples["input_ids"]):
                ch.append(token_id)
                if len(ch) == ch_len:
                    chs.append(ch)
                    ch = []
                    ch_len = chunk_len()

            if not chs:
                return []

            examples["input_ids"] = chs
            return examples

        return dataset.map(
            tokenize,
            batched=True,
            remove_columns=["text", "timestamp", "url"],
        ).map(chunk, batched=True, batch_size=self.hparams.concat_batch_size)

    def setup(self, stage=None):
        ds_train = self._create_dataset(split="train")
        ds_valid = self._create_dataset(split="validation")

        # FIXME: set current epoch on iterable datasets

        self.ds_train = self._create_pipeline(ds_train, min_seq_len=self.hparams.min_seq_len)
        self.ds_valid = self._create_pipeline(ds_valid, min_seq_len=None)

    def train_dataloader(self):
        return DataLoader(
            self.ds_train,
            collate_fn=self.collator,
            batch_size=self.hparams.batch_size,
            num_workers=self.hparams.num_train_workers,
            pin_memory=self.hparams.pin_memory,
        )

    def val_dataloader(self):
        return DataLoader(
            self.ds_valid,
            collate_fn=self.collator,
            batch_size=self.hparams.batch_size,
            num_workers=self.hparams.num_valid_workers,
            pin_memory=self.hparams.pin_memory,
        )


class C4Collator:
    def __init__(self, tokenizer):
        self.tokenizer = tokenizer

    def __call__(self, batch):
        batch = self.tokenizer.pad(batch, return_attention_mask=True, return_tensors="pt")
        return (
            batch["input_ids"][..., 1:],
            batch["input_ids"][..., :-1],
            ~batch["attention_mask"][..., :-1].type(torch.bool),
        )



from importlib.resources import path
from typing import Any, Optional

# Import to register additional optimizers at CLI
import torch_optimizer  # noqa: F401

from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.cli import LightningCLI, LRSchedulerTypeUnion
from pytorch_lightning.strategies import DDPStrategy
from torch.optim import Optimizer


class CLI(LightningCLI):
    def __init__(self, model_class, run=True, **kwargs):
        with path("perceiver.scripts", "trainer.yaml") as p:
            trainer_defaults = {"default_config_files": [str(p)]}

        super().__init__(
            model_class,
            run=run,
            save_config_kwargs={"overwrite": True},
            parser_kwargs={"fit": trainer_defaults, "test": trainer_defaults, "validate": trainer_defaults},
            **kwargs
        )

    def instantiate_trainer(self, **kwargs: Any) -> Trainer:
        if self.subcommand:
            cfg = self.config_init[self.subcommand]
        else:
            cfg = self.config_init

        if cfg["trainer"]["strategy"] == "ddp_static_graph":
            cfg["trainer"]["strategy"] = DDPStrategy(static_graph=True, find_unused_parameters=False)

        return super().instantiate_trainer(**kwargs)

    @staticmethod
    def configure_optimizers(
        lightning_module: LightningModule, optimizer: Optimizer, lr_scheduler: Optional[LRSchedulerTypeUnion] = None
    ):

        if lr_scheduler is None:
            return optimizer
        else:
            return {
                "optimizer": optimizer,
                "lr_scheduler": {"scheduler": lr_scheduler, "interval": "step", "frequency": 1},
            }



import string

from transformers import PerceiverTokenizer


class PerceiverTokenizerUtil:
    def __init__(self, tokenizer: PerceiverTokenizer):
        assert isinstance(tokenizer, PerceiverTokenizer)
        self.tokenizer = tokenizer
        self.whitespace_ids = set(tokenizer(string.whitespace, add_special_tokens=False).input_ids)

    def word_ids(self, token_ids):
        """Creates word ids from `token_ids`.

        Words boundaries are defined using whitespace boundaries. Whitespaces preceding a word have the same word id as
        the actual word following these whitespaces. Special tokens are assigned a `None` word id. Consecutive words do
        not necessarily have consecutive word ids. This method only guarantees that distinct words have distinct word
        ids. This is sufficient for `WordMaskingCollator` to function properly.
        """
        word_ids = []
        curr_id = 0

        special_mask = self.tokenizer.get_special_tokens_mask(token_ids, already_has_special_tokens=True)
        regular_token = True

        for i, token_id in enumerate(token_ids):
            if special_mask[i]:
                word_ids.append(None)
                curr_id += 1
            elif token_id in self.whitespace_ids:
                if regular_token:
                    regular_token = False
                    curr_id += 1
                word_ids.append(curr_id)
            else:
                regular_token = True
                word_ids.append(curr_id)

        return word_ids



import math
from typing import Optional, Tuple

import torch
import torch.nn as nn
from einops import rearrange, repeat


def positions(b, n, shift: Optional[torch.Tensor] = None, device: Optional[torch.device] = None):
    pos = repeat(torch.arange(n, device=device), "n -> b n", b=b)

    if shift is not None:
        if shift.shape != (b, 1):
            raise ValueError(f"shift must have shape {(1, b)} but has shape {shift.shape}")
        pos = pos - shift

    return torch.clamp(pos, min=0)


class RotaryPositionEmbedding:
    # Specified in https://arxiv.org/abs/2104.09864
    # Modified from https://github.com/lucidrains/rotary-embedding-torch
    def __init__(self, frq_pos_enc: torch.Tensor, right_align: bool = False):
        # frq_pos_enc shape is (b, n, c).
        # frq_pos_enc is broadcast to (b, h, n, c).
        self.frq_pos_enc = rearrange(frq_pos_enc, "b n c -> b 1 n c")
        self.rotate_dim = frq_pos_enc.shape[-1]
        self.right_align = right_align

    def rotate(self, t):
        seq_len = t.shape[-2]
        if self.right_align:
            # q and k are right-aligned in Perceiver AR
            pos_enc = self.frq_pos_enc[..., -seq_len:, :]
        else:
            # q and k are left-aligned
            pos_enc = self.frq_pos_enc[..., :seq_len, :]

        t_rot, t_pass = t[..., : self.rotate_dim], t[..., self.rotate_dim :]
        t_rot = (t_rot * pos_enc.cos()) + (self._rotate_half(t_rot) * pos_enc.sin())

        return torch.cat((t_rot, t_pass), dim=-1)

    @staticmethod
    def _rotate_half(x):
        # Rearranges channel dimension [x1, x2, x3, x4, ...] -> [-x2, x1, -x4, x3, ...]
        x = rearrange(x, "... (c r) -> ... c r", r=2)
        x1, x2 = x.unbind(dim=-1)
        x = torch.stack((-x2, x1), dim=-1)
        return rearrange(x, "... c r -> ... (c r)")


class FrequencyPositionEncoding(nn.Module):
    """Encodes positions as inverse frequencies.

    ```
    inverse_frequencies = 10000 ** -2(i−1)/dim, i = 1, 2, ... dim // 2
    position_encodings = p * inverse_frequencies, p = 0, 1, ... seq_len - 1
    ```
    """

    def __init__(self, dim):
        super().__init__()
        inv_freq = 1.0 / (10000 ** (torch.arange(0, dim, 2).float() / dim))
        self.register_buffer("inv_freq", inv_freq)

    def forward(self, abs_pos):
        # outer product of absolute positions and inverse frequencies
        pos_enc = torch.einsum("b n, f -> b n f", abs_pos, self.inv_freq)
        # frequency position encodings (per example in batch b)
        return repeat(pos_enc, "... pf -> ... (pf r)", r=2)


class FourierPositionEncoding(nn.Module):
    def __init__(self, input_shape: Tuple[int, ...], num_frequency_bands: int):
        super().__init__()

        self.input_shape = input_shape
        self.num_frequency_bands = num_frequency_bands

        # create encodings for single example
        pos = self._positions()
        enc = self._position_encodings(pos)

        # flatten encodings along spatial dimensions
        enc = rearrange(enc, "... c -> (...) c")

        # position encoding prototype
        self.register_buffer("position_encoding", enc)

    def _positions(self, v_min=-1.0, v_max=1.0):
        """Create evenly spaced position coordinates for self.spatial_shape with values in [v_min, v_max].

        :param v_min: minimum coordinate value per dimension.
        :param v_max: maximum coordinate value per dimension.
        :return: position coordinates tensor of shape (*shape, len(shape)).
        """
        coords = [torch.linspace(v_min, v_max, steps=s) for s in self.input_shape]
        return torch.stack(torch.meshgrid(*coords), dim=len(self.input_shape))

    def _position_encodings(
        self, p: torch.Tensor, max_frequencies: Optional[Tuple[int, ...]] = None, include_positions: bool = True
    ) -> torch.Tensor:
        """Fourier-encode positions p using self.num_bands frequency bands.

        :param p: positions of shape (*d, c) where c = len(d).
        :param max_frequencies: maximum frequency for each dimension (1-tuple for sequences,
               2-tuple for images, ...). If `None` values are derived from shape of p.
        :param include_positions: whether to include input positions p in returned encodings tensor.
        :returns: position encodings tensor of shape (*d, c * (2 * num_bands + include_positions)).
        """
        encodings = []

        if max_frequencies is None:
            max_frequencies = p.shape[:-1]

        frequencies = [
            torch.linspace(1.0, max_freq / 2.0, self.num_frequency_bands, device=p.device)
            for max_freq in max_frequencies
        ]
        frequency_grids = []

        for i, frequencies_i in enumerate(frequencies):
            frequency_grids.append(p[..., i : i + 1] * frequencies_i[None, ...])

        if include_positions:
            encodings.append(p)

        encodings.extend([torch.sin(math.pi * frequency_grid) for frequency_grid in frequency_grids])
        encodings.extend([torch.cos(math.pi * frequency_grid) for frequency_grid in frequency_grids])

        return torch.cat(encodings, dim=-1)

    def num_position_encoding_channels(self, include_positions: bool = True) -> int:
        return len(self.input_shape) * (2 * self.num_frequency_bands + include_positions)

    def forward(self, b):
        return repeat(self.position_encoding, "... -> b ...", b=b)



from perceiver.model.core.adapter import *
from perceiver.model.core.config import *
from perceiver.model.core.lightning import *
from perceiver.model.core.modules import *
from perceiver.model.core.position import *



import hashlib
import os
from enum import Enum
from itertools import chain
from typing import Optional, Sequence

import pytorch_lightning as pl
import torch
from datasets import Dataset, DatasetDict
from torch.utils.data import DataLoader
from transformers import AutoTokenizer

from perceiver.data.text.collator import DefaultCollator, TokenMaskingCollator, WordMaskingCollator
from perceiver.data.text.utils import PerceiverTokenizerUtil


os.environ["TOKENIZERS_PARALLELISM"] = "false"


class TextPreprocessor:
    def __init__(self, tokenizer: str, max_seq_len: int, add_special_tokens: bool):
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer, verbose=False)
        self.max_seq_len = max_seq_len
        self.add_special_tokens = add_special_tokens

    def preprocess(self, text):
        xs, pad_mask = self.preprocess_batch([text])
        return xs[0], pad_mask[0]

    def preprocess_batch(self, text_batch):
        result = self.tokenizer(
            text_batch,
            padding=self.tokenizer.pad_token is not None,
            truncation=True,
            add_special_tokens=self.add_special_tokens,
            return_token_type_ids=False,
            return_attention_mask=True,
            max_length=self.max_seq_len,
            return_tensors="pt",
        )
        return result["input_ids"], ~result["attention_mask"].type(torch.bool)


class Task(Enum):
    mlm = 0
    clm = 1
    clf = 2


class TextDataModule(pl.LightningDataModule):
    def __init__(
        self,
        dataset_dir: str,
        tokenizer: str,
        max_seq_len: int,
        task: Task = Task.mlm,
        mask_prob: float = 0.15,
        mask_words: bool = True,
        static_masking: bool = False,
        add_special_tokens: bool = False,
        add_eos_token: bool = False,
        padding_side: Optional[str] = None,
        random_train_shift: bool = False,
        random_valid_shift: bool = False,
        random_train_truncation: bool = False,
        random_valid_truncation: bool = False,
        random_min_seq_len: int = 16,
        preproc_batch_size: int = 1000,
        preproc_workers: Optional[int] = None,
        batch_size: int = 64,
        valid_batch_size: Optional[int] = None,
        num_workers: int = 3,
        pin_memory: bool = True,
    ):
        """Base class for consistent data preprocessing and loading across different text data sources.

        :param dataset_dir: Directory for storing the preprocessed dataset.
        :param tokenizer: Reference to a Hugging Face fast tokenizer (or the `deepmind/language-perceiver` tokenizer).
        :param max_seq_len: Maximum sequence length generated by this data module.
        :param task: The task for which this data module is used. Data are preprocessed and loaded in a task-specific
            way.
        :param mask_prob: Masking probability. Ignored if task is not `Task.mlm`.
        :param mask_words: Whether to mask words or individual tokens. Ignored if task is not `Task.mlm`.
        :param static_masking: Whether to mask at preprocessing time (static) or at data loading time (dynamic).
            Ignored if task is not `Task.mlm`.
        :param add_special_tokens: Whether to add special tokens to tokenized text.
        :param add_eos_token: Whether to append an EOS tokens to each example.
        :param padding_side: If `None`, uses the pre-configured `padding_side` of the tokenizer. Can be overridden by
            setting to "left" or "right".
        :param random_train_truncation: Randomly truncates sequences in the training set to length
            `randint(random_min_seq_len, max_seq_len + 1)`.
        :param random_valid_truncation: Randomly truncates sequences in the validation set to length
            `randint(random_min_seq_len, max_seq_len + 1)`.
        :param random_min_seq_len: Minimum sequence length when using `random_train_truncation` or
            `random_valid_truncation`.
        :param preproc_batch_size: Preprocessing batch size.
        :param preproc_workers: Number of preprocessing processes. If not defined, defaults to `num_workers`.
        :param batch_size: Batch size of loaded training data.
        :param valid_batch_size: Batch size of loaded validation data. If `None` defaults to `batch_size`
        :param num_workers: Number of data loading processes.
        """

        super().__init__()
        self.save_hyperparameters()

        if self.hparams.static_masking and not self.hparams.mask_words:
            raise ValueError("static_masking=true is only supported for mask_words=true")

        self.tokenizer = AutoTokenizer.from_pretrained(self.hparams.tokenizer, verbose=False)

        if self.hparams.padding_side is not None:
            self.tokenizer.padding_side = self.hparams.padding_side

        # PerceiverTokenizer needs special support for generating word_ids as it is not a fast tokenizer
        self.perceiver_tokenizer_configured = self.hparams.tokenizer == "deepmind/language-perceiver"
        if self.perceiver_tokenizer_configured:
            self.perceiver_tokenizer_util = PerceiverTokenizerUtil(self.tokenizer)

        if self.hparams.task == Task.mlm and not self.hparams.static_masking:
            if self.hparams.mask_words:
                self.collator = WordMaskingCollator(tokenizer=self.tokenizer, mask_prob=self.hparams.mask_prob)
            else:
                self.collator = TokenMaskingCollator(tokenizer=self.tokenizer, mask_prob=self.hparams.mask_prob)
        else:
            self.collator = DefaultCollator(tokenizer=self.tokenizer, max_seq_len=self.hparams.max_seq_len)

        self.ds_train = None
        self.ds_valid = None

    @property
    def valid_batch_size(self):
        if self.hparams.valid_batch_size is None:
            return self.hparams.batch_size
        else:
            return self.hparams.valid_batch_size

    @property
    def vocab_size(self):
        return self.tokenizer.vocab_size

    @property
    def max_seq_len(self):
        return self.hparams.max_seq_len

    @property
    def random_shift(self):
        return self.hparams.random_train_shift or self.hparams.random_valid_shift

    @property
    def preproc_workers(self):
        if self.hparams.preproc_workers is not None:
            return self.hparams.preproc_workers
        else:
            return max(1, self.hparams.num_workers)

    @property
    def preproc_dir(self):
        h = hashlib.new("md5")
        h.update(self.preproc_dir_hash_input().encode())
        return os.path.join(self.hparams.dataset_dir, "preproc", h.hexdigest())

    def preproc_dir_hash_input(self) -> str:
        hash_input = f"{self.hparams.tokenizer}-{self.max_seq_len}-{self.hparams.task.name}-{self.random_shift}"
        if self.hparams.task == Task.mlm and self.hparams.static_masking:
            hash_input = f"{hash_input}-{self.hparams.mask_words}-{self.hparams.mask_prob}"
        if self.hparams.add_special_tokens:
            hash_input = f"{hash_input}-st"
        if self.hparams.add_eos_token:
            hash_input = f"{hash_input}-eos"
        if self.hparams.get("source_train_size") is not None:
            hash_input = f"{hash_input}-ts-{self.hparams.source_train_size}"
        if self.hparams.get("source_valid_size") is not None:
            hash_input = f"{hash_input}-vs-{self.hparams.source_valid_size}"
        return hash_input

    def prepare_data(self) -> None:
        if not os.path.exists(self.preproc_dir):
            dataset = self.load_source_dataset()
            dataset = self._prepare_dataset(dataset)
            dataset.save_to_disk(self.preproc_dir)

    def setup(self, stage=None):
        dataset = self.load_prepared_dataset()

        self.ds_train = dataset["train"]
        self.ds_valid = dataset["valid"]

        if self.hparams.task in [Task.clm, Task.mlm]:
            if self.hparams.random_train_shift:
                self.ds_train = RandomShiftDataset(self.ds_train)
            if self.hparams.random_valid_shift:
                self.ds_valid = RandomShiftDataset(self.ds_valid)

        if self.hparams.random_train_truncation:
            self.ds_train = RandomTruncationDataset(self.ds_train, self.hparams.random_min_seq_len)
        if self.hparams.random_valid_truncation:
            self.ds_valid = RandomTruncationDataset(self.ds_valid, self.hparams.random_min_seq_len)

        if self.hparams.task == Task.clm:
            self.ds_train = CLMDataset(self.ds_train)
            self.ds_valid = CLMDataset(self.ds_valid)

    def train_dataloader(self):
        return DataLoader(
            self.ds_train,
            shuffle=True,
            collate_fn=self.collator,
            batch_size=self.hparams.batch_size,
            num_workers=self.hparams.num_workers,
            pin_memory=self.hparams.pin_memory,
        )

    def val_dataloader(self):
        return DataLoader(
            self.ds_valid,
            shuffle=False,
            collate_fn=self.collator,
            batch_size=self.valid_batch_size,
            num_workers=self.hparams.num_workers,
            pin_memory=self.hparams.pin_memory,
        )

    def text_preprocessor(self):
        preproc = TextPreprocessor(
            tokenizer=self.hparams.tokenizer,
            max_seq_len=self.hparams.max_seq_len,
            add_special_tokens=self.hparams.add_special_tokens,
        )

        if self.hparams.padding_side is not None:
            preproc.tokenizer.padding_side = self.hparams.padding_side

        return preproc

    def load_source_dataset(self) -> DatasetDict:
        """Must return a DatasetDict with keys 'train' and 'valid'."""
        raise NotImplementedError()

    def load_prepared_dataset(self) -> DatasetDict:
        return DatasetDict.load_from_disk(self.preproc_dir)

    def _prepare_dataset(self, dataset: DatasetDict):
        if self.hparams.task == Task.clm:
            dataset = self._tokenize_dataset(dataset, return_word_ids=False)
            dataset = self._chunk_dataset(dataset, chunk_size=self.hparams.max_seq_len + 1, include_keys=["input_ids"])
        elif self.hparams.task == Task.mlm:
            dataset = self._tokenize_dataset(dataset, return_word_ids=True)
            dataset = self._chunk_dataset(dataset, chunk_size=self.hparams.max_seq_len)
            if self.hparams.static_masking:
                dataset = self._mask_dataset(dataset)
        else:  # task == Task.clf
            assert "label" in dataset["train"].column_names
            assert "label" in dataset["valid"].column_names
            dataset = self._tokenize_dataset(
                dataset, max_length=self.max_seq_len, truncation=True, return_word_ids=False
            )
        return dataset

    def _tokenize_dataset(
        self,
        dataset: DatasetDict,
        padding=False,
        truncation=False,
        max_length=None,
        return_word_ids=True,
    ):
        def tokenize(examples):
            if self.hparams.add_eos_token:
                examples["text"] = [text + self.tokenizer.eos_token for text in examples["text"]]
            encoding = self.tokenizer(
                examples["text"],
                padding=padding,
                truncation=truncation,
                max_length=max_length,
                add_special_tokens=self.hparams.add_special_tokens,
                return_token_type_ids=False,
                return_attention_mask=False,
            )
            if return_word_ids:
                if self.perceiver_tokenizer_configured:
                    encoding["word_ids"] = [
                        self.perceiver_tokenizer_util.word_ids(input_ids) for input_ids in encoding["input_ids"]
                    ]
                else:
                    encoding["word_ids"] = [encoding.word_ids(i) for i in range(len(encoding["input_ids"]))]
            return encoding

        result = DatasetDict()
        for key in dataset.keys():
            result[key] = dataset[key].map(
                tokenize,
                batched=True,
                batch_size=self.hparams.preproc_batch_size,
                num_proc=self.preproc_workers,
                remove_columns=["text"],
                load_from_cache_file=False,
                desc="Running tokenizer on dataset",
            )
        return result

    def _chunk_dataset(
        self,
        dataset: DatasetDict,
        chunk_size: int,
        include_keys: Sequence[str] = ("input_ids", "word_ids"),
        remove_keys: Sequence[str] = (),
    ):
        def chunk(*args):
            chained = {k: list(chain(*args[i])) for i, k in enumerate(include_keys)}
            chained_len = len(chained[include_keys[0]])
            if chained_len >= chunk_size:
                chained_len = (chained_len // chunk_size) * chunk_size
            return {k: [t[i : i + chunk_size] for i in range(0, chained_len, chunk_size)] for k, t in chained.items()}

        result = DatasetDict()
        for key in dataset.keys():
            result[key] = dataset[key].map(
                chunk,
                batched=True,
                batch_size=self.hparams.preproc_batch_size,
                num_proc=self.preproc_workers,
                input_columns=list(include_keys),
                remove_columns=list(remove_keys),
                load_from_cache_file=False,
                desc=f"Split dataset into chunks of size {chunk_size}",
            )
        return result

    def _mask_dataset(self, dataset: DatasetDict):
        wmc = WordMaskingCollator(tokenizer=self.tokenizer, mask_prob=self.hparams.mask_prob)

        def mask(example):
            return wmc.mask_words_1(example)

        result = DatasetDict()
        for key in dataset.keys():
            result[key] = dataset[key].map(
                mask,
                batched=False,
                num_proc=self.preproc_workers,
                load_from_cache_file=False,
                desc="Mask words in dataset",
            )
        return result

    def _train_valid_split(self, dataset: Dataset, train_size, test_size):
        dataset = dataset.train_test_split(train_size=train_size, test_size=test_size, shuffle=not self.random_shift)
        return DatasetDict(train=dataset["train"], valid=dataset["test"])


class RandomShiftDataset(torch.utils.data.Dataset):
    def __init__(self, dataset):
        self.dataset = dataset

    def __getitem__(self, idx):
        example_1 = self.dataset[idx]
        example_2 = self.dataset[idx + 1]

        result = {}
        shift = None

        for key in example_1.keys():
            record_1 = example_1[key]
            record_2 = example_2[key]

            if shift is None:
                shift = torch.randint(len(record_1), (1,)).item()

            result[key] = record_1[shift:] + record_2[:shift]

        return result

    def __len__(self):
        return len(self.dataset) - 1


class RandomTruncationDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, random_min_seq_len: int):
        self.dataset = dataset
        self.random_min_seq_len = random_min_seq_len

    def __getitem__(self, idx):
        example = self.dataset[idx]
        example_seq_len = len(example["input_ids"])

        drop_max = example_seq_len - self.random_min_seq_len
        if drop_max > 0:
            drop = torch.randint(drop_max + 1, size=(1,))
            if drop > 0:
                for key in example.keys():
                    example[key] = example[key][:-drop]

        return example

    def __len__(self):
        return len(self.dataset)


class CLMDataset(torch.utils.data.Dataset):
    def __init__(self, dataset):
        self.dataset = dataset

    def __getitem__(self, idx):
        record = self.dataset[idx]["input_ids"]
        return {"input_ids": record[:-1], "label_ids": record[1:]}

    def __len__(self):
        return len(self.dataset)



from typing import Any, Optional

import pytorch_lightning as pl

from perceiver.model.core.config import DecoderConfig, EncoderConfig, PerceiverIOConfig


class LitPerceiverIO(pl.LightningModule):
    def __init__(
        self,
        encoder: EncoderConfig,
        decoder: DecoderConfig,
        num_latents: int,
        num_latent_channels: int,
        activation_checkpointing: bool = False,
        activation_offloading: bool = False,
        params: Optional[str] = None,
    ):
        super().__init__()
        self.save_hyperparameters()

    @classmethod
    def create(cls, config: PerceiverIOConfig, *args: Any, **kwargs: Any):
        return cls(
            config.encoder,
            config.decoder,
            *args,
            num_latents=config.num_latents,
            num_latent_channels=config.num_latent_channels,
            activation_checkpointing=config.activation_checkpointing,
            activation_offloading=config.activation_offloading,
            params=config.params,
            **kwargs,
        )

    @classmethod
    def load_from_checkpoint(cls, *args, params=None, **kwargs: Any):
        return super().load_from_checkpoint(*args, params=params, **kwargs)



from pytorch_lightning.cli import LightningArgumentParser

from perceiver.model.vision.image_classifier import LitImageClassifier
from perceiver.scripts.cli import CLI
from perceiver.scripts.lrs import ConstantWithWarmupLR


class ImageClassifierCLI(CLI):
    def add_arguments_to_parser(self, parser: LightningArgumentParser) -> None:
        super().add_arguments_to_parser(parser)
        parser.add_lr_scheduler_args(ConstantWithWarmupLR)
        parser.link_arguments("data.image_shape", "model.encoder.image_shape", apply_on="instantiate")
        parser.link_arguments("data.num_classes", "model.decoder.num_classes", apply_on="instantiate")
        parser.set_defaults(
            {
                "model.num_latents": 512,
                "model.num_latent_channels": 1024,
                "model.encoder.num_frequency_bands": 64,
                "model.encoder.num_cross_attention_layers": 1,
                "model.encoder.num_cross_attention_heads": 1,
                "model.encoder.num_self_attention_heads": 8,
                "model.encoder.num_self_attention_layers_per_block": 6,
                "model.encoder.num_self_attention_blocks": 8,
                "model.encoder.dropout": 0.1,
                "model.decoder.num_output_query_channels": 1024,
                "model.decoder.num_cross_attention_heads": 1,
                "model.decoder.dropout": 0.1,
            }
        )


if __name__ == "__main__":
    ImageClassifierCLI(LitImageClassifier, run=True)



import multiprocessing as mp

import jsonargparse

from perceiver.data.text import (
    BookCorpusDataModule,
    BookCorpusOpenDataModule,
    Enwik8DataModule,
    ImdbDataModule,
    Task,
    WikipediaDataModule,
    WikiTextDataModule,
)


DATAMODULE_CLASSES = {
    "bookcorpus": BookCorpusDataModule,
    "bookcorpusopen": BookCorpusOpenDataModule,
    "enwik8": Enwik8DataModule,
    "imdb": ImdbDataModule,
    "wikipedia": WikipediaDataModule,
    "wikitext": WikiTextDataModule,
}


def main(args):
    dm_class = DATAMODULE_CLASSES[args.dataset]

    del args.dataset

    dm_class(**args, pin_memory=False).prepare_data()


if __name__ == "__main__":
    parser = jsonargparse.ArgumentParser(description="Preprocess dataset for training")
    parser.add_argument("dataset", default="wikitext", choices=list(DATAMODULE_CLASSES.keys()))
    parser.add_argument("--tokenizer", type=str)
    parser.add_argument("--max_seq_len", type=int)
    parser.add_argument("--task", default=Task.mlm, type=Task)
    parser.add_argument("--mask_prob", default=0.15, type=float)
    parser.add_argument("--mask_words", default=True, type=bool)
    parser.add_argument("--static_masking", default=False, type=bool)
    parser.add_argument("--add_special_tokens", default=False, type=bool)
    parser.add_argument("--add_eos_token", default=False, type=bool)
    parser.add_argument("--random_train_shift", default=False, type=bool)
    parser.add_argument("--preproc_workers", default=mp.cpu_count(), type=int)
    main(parser.parse_args())



from PIL import Image
from torchvision.transforms import Compose
from transformers.models.perceiver.feature_extraction_perceiver import PerceiverFeatureExtractor

from perceiver.data.vision.common import channels_to_last, ImagePreprocessor


class ImageNetPreprocessor(ImagePreprocessor):
    def __init__(self, crop_size: int = 256, size: int = 224, channels_last: bool = True):
        super().__init__(imagenet_valid_transform(crop_size, size, channels_last))


def imagenet_valid_transform(crop_size: int, size: int, channels_last: bool):
    extractor = PerceiverFeatureExtractor(
        do_center_crop=True,
        crop_size=crop_size,
        do_resize=True,
        size=size,
        resample=Image.BICUBIC,
        do_normalize=True,
    )

    def extract(img):
        return extractor(img, return_tensors="pt")["pixel_values"][0]

    transforms = [extract]

    if channels_last:
        transforms.append(channels_to_last)

    return Compose(transforms)



import os
from typing import Optional

import pytorch_lightning as pl
from datasets import load_dataset
from torch.utils.data import DataLoader
from torchvision import transforms

from perceiver.data.vision.common import channels_to_last, ImagePreprocessor, lift_transform


class MNISTPreprocessor(ImagePreprocessor):
    def __init__(self, normalize: bool = True, channels_last: bool = True):
        super().__init__(mnist_transform(normalize, channels_last))


class MNISTDataModule(pl.LightningDataModule):
    def __init__(
        self,
        dataset_dir: str = os.path.join(".cache", "mnist"),
        normalize: bool = True,
        channels_last: bool = True,
        random_crop: Optional[int] = None,
        batch_size: int = 64,
        num_workers: int = 3,
        pin_memory: bool = True,
        shuffle: bool = True,
    ):
        super().__init__()
        self.save_hyperparameters()
        self.channels_last = channels_last

        self.tf_train = mnist_transform(normalize, channels_last, random_crop=random_crop)
        self.tf_valid = mnist_transform(normalize, channels_last, random_crop=None)

        self.ds_train = None
        self.ds_valid = None

    @property
    def num_classes(self):
        return 10

    @property
    def image_shape(self):
        if self.hparams.channels_last:
            return 28, 28, 1
        else:
            return 1, 28, 28

    def load_dataset(self, split: Optional[str] = None):
        return load_dataset("mnist", split=split, cache_dir=self.hparams.dataset_dir)

    def prepare_data(self) -> None:
        self.load_dataset()

    def setup(self, stage: Optional[str] = None) -> None:
        self.ds_train = self.load_dataset(split="train")
        self.ds_train.set_transform(lift_transform(self.tf_train))

        self.ds_valid = self.load_dataset(split="test")
        self.ds_valid.set_transform(lift_transform(self.tf_valid))

    def train_dataloader(self):
        return DataLoader(
            self.ds_train,
            shuffle=self.hparams.shuffle,
            batch_size=self.hparams.batch_size,
            num_workers=self.hparams.num_workers,
            pin_memory=self.hparams.pin_memory,
        )

    def val_dataloader(self):
        return DataLoader(
            self.ds_valid,
            shuffle=False,
            batch_size=self.hparams.batch_size,
            num_workers=self.hparams.num_workers,
            pin_memory=self.hparams.pin_memory,
        )


def mnist_transform(normalize: bool = True, channels_last: bool = True, random_crop: Optional[int] = None):
    transform_list = []

    if random_crop is not None:
        transform_list.append(transforms.RandomCrop(random_crop))

    transform_list.append(transforms.ToTensor())

    if normalize:
        transform_list.append(transforms.Normalize(mean=(0.5,), std=(0.5,)))

    if channels_last:
        transform_list.append(channels_to_last)

    return transforms.Compose(transform_list)



import os
from typing import Any

from perceiver.model.core import PerceiverIO, PerceiverIOConfig, TrainableQueryProvider
from perceiver.model.core.classifier import (
    ClassificationDecoder,
    ClassificationDecoderConfig,
    ClassificationOutputAdapter,
    LitClassifier,
)
from perceiver.model.core.utils import is_checkpoint
from perceiver.model.text.common import TextEncoder, TextEncoderConfig
from perceiver.model.text.mlm import LitMaskedLanguageModel


class TextClassifier(PerceiverIO):
    def __init__(self, config: PerceiverIOConfig[TextEncoderConfig, ClassificationDecoderConfig]):
        encoder = TextEncoder(
            config.encoder,
            num_latents=config.num_latents,
            num_latent_channels=config.num_latent_channels,
            activation_checkpointing=config.activation_checkpointing,
            activation_offloading=config.activation_offloading,
        )
        output_query_provider = TrainableQueryProvider(
            num_queries=config.decoder.num_output_queries,
            num_query_channels=config.decoder.num_output_query_channels,
            init_scale=config.decoder.init_scale,
        )
        output_adapter = ClassificationOutputAdapter(
            num_classes=config.decoder.num_classes,
            num_output_query_channels=config.decoder.num_output_query_channels,
        )
        decoder = ClassificationDecoder(
            output_adapter=output_adapter,
            output_query_provider=output_query_provider,
            num_latent_channels=config.num_latent_channels,
            activation_checkpointing=config.activation_checkpointing,
            activation_offloading=config.activation_offloading,
            **config.decoder.base_kwargs()
        )
        super().__init__(encoder, decoder)


class LitTextClassifier(LitClassifier):
    def __init__(self, encoder: TextEncoderConfig, decoder: ClassificationDecoderConfig, *args: Any, **kwargs: Any):
        super().__init__(encoder, decoder, *args, **kwargs)
        self.model = TextClassifier(
            PerceiverIOConfig(
                encoder=encoder,
                decoder=decoder,
                num_latents=self.hparams.num_latents,
                num_latent_channels=self.hparams.num_latent_channels,
                activation_checkpointing=self.hparams.activation_checkpointing,
                activation_offloading=self.hparams.activation_offloading,
                params=self.hparams.params,
            )
        )

        model_params = self.hparams.params
        encoder_params = self.hparams.encoder.params

        if model_params is not None and is_checkpoint(model_params):
            lit_model = LitTextClassifier.load_from_checkpoint(model_params, params=None)
            self.model.load_state_dict(lit_model.model.state_dict())
        if encoder_params is not None and is_checkpoint(encoder_params) and os.path.exists(encoder_params):
            lit_model = LitMaskedLanguageModel.load_from_checkpoint(encoder_params, params=None)
            self.model.encoder.load_state_dict(lit_model.model.encoder.state_dict())

    def step(self, batch):
        y, x, x_mask = batch
        return self.loss_acc(self(x, x_mask), y)

    def forward(self, x, x_mask=None):
        return self.model(x, x_mask)



from perceiver.data.text.bookcorpus import BookCorpusDataModule
from perceiver.data.text.bookcorpusopen import BookCorpusOpenDataModule
from perceiver.data.text.c4 import C4DataModule
from perceiver.data.text.common import Task, TextPreprocessor
from perceiver.data.text.enwik8 import Enwik8DataModule
from perceiver.data.text.imdb import ImdbDataModule
from perceiver.data.text.wikipedia import WikipediaDataModule
from perceiver.data.text.wikitext import WikiTextDataModule



from typing import Sequence

import torch
import torch.nn as nn
import transformers

from perceiver.model.core.modules import (
    CrossAttentionLayer,
    MLP,
    MultiHeadAttention,
    PerceiverEncoder,
    SelfAttentionLayer,
)


def copy_param(src: nn.Parameter, tgt: nn.Parameter):
    with torch.no_grad():
        tgt.copy_(src)


def copy_params(src: nn.Module, tgt: nn.Module):
    tgt.load_state_dict(src.state_dict())


def copy_attention_params(src: transformers.PerceiverLayer, tgt: MultiHeadAttention):
    copy_params(src.attention.self.query, tgt.q_proj)
    copy_params(src.attention.self.key, tgt.k_proj)
    copy_params(src.attention.self.value, tgt.v_proj)
    copy_params(src.attention.output.dense, tgt.o_proj)


def copy_mlp_params(src: transformers.PerceiverLayer, tgt: MLP):
    copy_params(src.layernorm, tgt[0])
    copy_params(src.mlp.dense1, tgt[1])
    copy_params(src.mlp.dense2, tgt[3])


def copy_cross_attention_layer_params(src: transformers.PerceiverLayer, tgt: CrossAttentionLayer, query_residual: bool):
    att_tgt = tgt[0].module if query_residual else tgt[0]
    mlp_tgt = tgt[1].module

    copy_params(src.attention.self.layernorm1, att_tgt.q_norm)
    copy_params(src.attention.self.layernorm2, att_tgt.kv_norm)

    copy_attention_params(src, att_tgt.attention)
    copy_mlp_params(src, mlp_tgt)


def copy_self_attention_layer_params(src: transformers.PerceiverLayer, tgt: SelfAttentionLayer):
    att_tgt = tgt[0].module
    mlp_tgt = tgt[1].module

    copy_params(src.attention.self.layernorm1, att_tgt.norm)

    copy_attention_params(src, att_tgt.attention)
    copy_mlp_params(src, mlp_tgt)


def copy_self_attention_block_params(src: Sequence[transformers.PerceiverLayer], tgt: Sequence[SelfAttentionLayer]):
    assert len(src) == len(tgt)

    for src_layer, tgt_layer in zip(src, tgt):
        copy_self_attention_layer_params(src_layer, tgt_layer)


def copy_latent_provider_params(src: transformers.PerceiverModel, tgt: PerceiverEncoder):
    copy_param(src.embeddings.latents, tgt.latent_provider._query)



from typing import Optional

import torch
import torch.nn as nn
from einops import rearrange
from fairscale.nn import checkpoint_wrapper

from perceiver.model.core.adapter import (
    InputAdapter,
    OutputAdapter,
    QueryProvider,
    RotarySupport,
    TrainableQueryProvider,
)
from perceiver.model.core.position import positions, RotaryPositionEmbedding
from perceiver.model.core.utils import init_parameters, Residual, Sequential


class MultiHeadAttention(nn.Module):
    def __init__(
        self,
        num_heads: int,
        num_q_input_channels: int,
        num_kv_input_channels: int,
        num_qk_channels: Optional[int] = None,
        num_v_channels: Optional[int] = None,
        num_output_channels: Optional[int] = None,
        max_heads_parallel: Optional[int] = None,
        causal_attention: bool = False,
        dropout: float = 0.0,
        qkv_bias: bool = True,
        out_bias: bool = True,
    ):
        """Multi-head attention as specified in https://arxiv.org/abs/2107.14795 Appendix E plus support for rotary
        position embeddings (https://arxiv.org/abs/2104.09864) and causal attention. Causal attention requires
        queries and keys to be right-aligned, if they have different length.

        :param num_heads: Number of attention heads.
        :param num_q_input_channels: Number of query input channels.
        :param num_kv_input_channels: Number of key/value input channels.
        :param num_qk_channels: Number of query and key channels. Default is number `num_q_input_channels`
        :param num_v_channels: Number of value channels. Default is `num_qk_channels`.
        :param num_output_channels: Number of output channels. Default is `num_q_input_channels`
        :param max_heads_parallel: Maximum number of heads to be processed in parallel. Default is `num_heads`.
        :param causal_attention: Whether to apply a causal attention mask. Default is `False`.
        :param dropout: Dropout probability for attention matrix values. Default is `0.0`
        :param qkv_bias: Whether to use a bias term for query, key and value projections. Default is `True`.
        :param qkv_bias: Whether to use a bias term for output projection. Default is `True`.
        """
        super().__init__()

        if num_qk_channels is None:
            num_qk_channels = num_q_input_channels

        if num_v_channels is None:
            num_v_channels = num_qk_channels

        if num_output_channels is None:
            num_output_channels = num_q_input_channels

        if num_qk_channels % num_heads != 0:
            raise ValueError("num_qk_channels must be divisible by num_heads")

        if num_v_channels % num_heads != 0:
            raise ValueError("num_v_channels must be divisible by num_heads")

        num_qk_channels_per_head = num_qk_channels // num_heads

        self.dp_scale = num_qk_channels_per_head ** -0.5
        self.num_heads = num_heads
        self.causal_attention = causal_attention

        if max_heads_parallel is None:
            self.max_heads_parallel = num_heads
        else:
            self.max_heads_parallel = max_heads_parallel

        self.q_proj = nn.Linear(num_q_input_channels, num_qk_channels, bias=qkv_bias)
        self.k_proj = nn.Linear(num_kv_input_channels, num_qk_channels, bias=qkv_bias)
        self.v_proj = nn.Linear(num_kv_input_channels, num_v_channels, bias=qkv_bias)
        self.o_proj = nn.Linear(num_v_channels, num_output_channels, bias=out_bias)
        self.dropout = nn.Dropout(dropout)

    def forward(
        self,
        x_q,
        x_kv,
        pad_mask=None,
        rot_pos_emb_q: Optional[RotaryPositionEmbedding] = None,
        rot_pos_emb_k: Optional[RotaryPositionEmbedding] = None,
    ):
        """
        :param x_q: Query input of shape (B, N, D) where B is the batch size, N the query sequence length
            and D the number of query input channels (= `num_q_input_channels`)
        :param x_kv: Key/value input of shape (B, L, C) where B is the batch size, L the key/value sequence
            length and C are the number of key/value input channels (= `num_kv_input_channels`)
        :param pad_mask: Boolean key padding mask. `True` values indicate padding tokens.
        :param rot_pos_emb_q: Applies a rotary position embedding to query i.e. if defined, rotates the query.
        :param rot_pos_emb_k: Applies a rotary position embedding to key i.e. if defined, rotates the key.
        :return: attention result of shape (B, N, F) where B is the batch size, N the query sequence length
            and F the number of output channels (= `num_output_channels`)
        """

        q = self.q_proj(x_q)
        k = self.k_proj(x_kv)
        v = self.v_proj(x_kv)

        q, k, v = (rearrange(x, "b n (h c) -> b h n c", h=self.num_heads) for x in [q, k, v])
        q = q * self.dp_scale

        if rot_pos_emb_q is not None:
            q = rot_pos_emb_q.rotate(q)

        if rot_pos_emb_k is not None:
            k = rot_pos_emb_k.rotate(k)

        if pad_mask is not None:
            pad_mask = rearrange(pad_mask, "b j -> b 1 1 j")

        if self.causal_attention:
            i = q.shape[2]
            j = k.shape[2]

            # If q and k have different length, causal masking only works if they are right-aligned.
            causal_mask = torch.ones((i, j), device=x_q.device, dtype=torch.bool).triu(j - i + 1)

        o_chunks = []

        # Only process a given maximum number of heads in
        # parallel, using several iterations, if necessary.
        for q_chunk, k_chunk, v_chunk in zip(
            q.split(self.max_heads_parallel, dim=1),
            k.split(self.max_heads_parallel, dim=1),
            v.split(self.max_heads_parallel, dim=1),
        ):
            attn = torch.einsum("b h i c, b h j c -> b h i j", q_chunk, k_chunk)
            attn_max_neg = -torch.finfo(attn.dtype).max

            if pad_mask is not None:
                attn.masked_fill_(pad_mask, attn_max_neg)

            if self.causal_attention:
                attn.masked_fill_(causal_mask, attn_max_neg)

            attn = attn.softmax(dim=-1)
            attn = self.dropout(attn)

            o_chunk = torch.einsum("b h i j, b h j c -> b h i c", attn, v_chunk)
            o_chunks.append(o_chunk)

        o = torch.cat(o_chunks, dim=1)
        o = rearrange(o, "b h n c -> b n (h c)", h=self.num_heads)

        return self.o_proj(o)


class CrossAttention(nn.Module):
    def __init__(
        self,
        num_heads: int,
        num_q_input_channels: int,
        num_kv_input_channels: int,
        num_qk_channels: Optional[int] = None,
        num_v_channels: Optional[int] = None,
        max_heads_parallel: Optional[int] = None,
        causal_attention: bool = False,
        dropout: float = 0.0,
        qkv_bias: bool = True,
        out_bias: bool = True,
    ):
        """Pre-layer-norm cross-attention (see `MultiHeadAttention` for attention details)."""
        super().__init__()
        self.q_norm = nn.LayerNorm(num_q_input_channels)
        self.kv_norm = nn.LayerNorm(num_kv_input_channels)
        self.attention = MultiHeadAttention(
            num_heads=num_heads,
            num_q_input_channels=num_q_input_channels,
            num_kv_input_channels=num_kv_input_channels,
            num_qk_channels=num_qk_channels,
            num_v_channels=num_v_channels,
            max_heads_parallel=max_heads_parallel,
            causal_attention=causal_attention,
            dropout=dropout,
            qkv_bias=qkv_bias,
            out_bias=out_bias,
        )

    def forward(self, x_q, x_kv=None, x_kv_prefix=None, pad_mask=None, rot_pos_emb_q=None, rot_pos_emb_k=None):
        """Pre-layer-norm cross-attention of query input `x_q` to key/value input (`x_kv` or `x_kv_prefix`).

        If `x_kv_prefix` is defined, the entire key/value input is a concatenation of `x_kv_prefix` and `x_q` along
        the sequence dimension. In this case, the query attends to itself at the end of the key/value sequence (use
        case: Perceiver AR). If `x_kv_prefix` is not defined, `x_kv` is the entire key/value input.
        """
        x_q = self.q_norm(x_q)

        if x_kv is None:
            x_kv_prefix = self.kv_norm(x_kv_prefix)
            x_kv = torch.cat([x_kv_prefix, x_q], dim=1)
        else:
            x_kv = self.kv_norm(x_kv)

        return self.attention(x_q, x_kv, pad_mask=pad_mask, rot_pos_emb_q=rot_pos_emb_q, rot_pos_emb_k=rot_pos_emb_k)


class SelfAttention(nn.Module):
    def __init__(
        self,
        num_heads: int,
        num_channels: int,
        num_qk_channels: Optional[int] = None,
        num_v_channels: Optional[int] = None,
        max_heads_parallel: Optional[int] = None,
        causal_attention: bool = False,
        dropout: float = 0.0,
        qkv_bias: bool = True,
        out_bias: bool = True,
    ):
        """Pre-layer norm self-attention (see `MultiHeadAttention` and for attention details)."""
        super().__init__()
        self.norm = nn.LayerNorm(num_channels)
        self.attention = MultiHeadAttention(
            num_heads=num_heads,
            num_q_input_channels=num_channels,
            num_kv_input_channels=num_channels,
            num_qk_channels=num_qk_channels,
            num_v_channels=num_v_channels,
            max_heads_parallel=max_heads_parallel,
            causal_attention=causal_attention,
            dropout=dropout,
            qkv_bias=qkv_bias,
            out_bias=out_bias,
        )

    def forward(self, x, pad_mask=None, rot_pos_emb=None):
        """Pre-layer-norm self-attention of input `x`."""
        x = self.norm(x)
        return self.attention(x, x, pad_mask=pad_mask, rot_pos_emb_q=rot_pos_emb, rot_pos_emb_k=rot_pos_emb)


class CrossAttentionLayer(Sequential):
    def __init__(
        self,
        num_heads: int,
        num_q_input_channels: int,
        num_kv_input_channels: int,
        num_qk_channels: Optional[int] = None,
        num_v_channels: Optional[int] = None,
        max_heads_parallel: Optional[int] = None,
        causal_attention: bool = False,
        widening_factor: int = 1,
        dropout: float = 0.0,
        attention_residual: bool = True,
        qkv_bias: bool = True,
        out_bias: bool = True,
        mlp_bias: bool = True,
    ):
        cross_attn = CrossAttention(
            num_heads=num_heads,
            num_q_input_channels=num_q_input_channels,
            num_kv_input_channels=num_kv_input_channels,
            num_qk_channels=num_qk_channels,
            num_v_channels=num_v_channels,
            max_heads_parallel=max_heads_parallel,
            causal_attention=causal_attention,
            dropout=dropout,
            qkv_bias=qkv_bias,
            out_bias=out_bias,
        )
        super().__init__(
            Residual(cross_attn) if attention_residual else cross_attn,
            Residual(MLP(num_q_input_channels, widening_factor, bias=mlp_bias)),
        )


class SelfAttentionLayer(Sequential):
    def __init__(
        self,
        num_heads: int,
        num_channels: int,
        num_qk_channels: Optional[int] = None,
        num_v_channels: Optional[int] = None,
        max_heads_parallel: Optional[int] = None,
        causal_attention: bool = False,
        widening_factor: int = 1,
        dropout: float = 0.0,
        qkv_bias: bool = True,
        out_bias: bool = True,
        mlp_bias: bool = True,
    ):
        self_attn = SelfAttention(
            num_heads=num_heads,
            num_channels=num_channels,
            num_qk_channels=num_qk_channels,
            num_v_channels=num_v_channels,
            max_heads_parallel=max_heads_parallel,
            causal_attention=causal_attention,
            dropout=dropout,
            qkv_bias=qkv_bias,
            out_bias=out_bias,
        )
        super().__init__(
            Residual(self_attn),
            Residual(MLP(num_channels, widening_factor, bias=mlp_bias)),
        )


class SelfAttentionBlock(Sequential):
    def __init__(
        self,
        num_layers: int,
        num_heads: int,
        num_channels: int,
        num_qk_channels: Optional[int] = None,
        num_v_channels: Optional[int] = None,
        max_heads_parallel: Optional[int] = None,
        causal_attention: bool = False,
        widening_factor: int = 1,
        dropout: float = 0.0,
        activation_checkpointing: bool = False,
        activation_offloading: bool = False,
        qkv_bias: bool = True,
        out_bias: bool = True,
        mlp_bias: bool = True,
    ):
        layers = [
            SelfAttentionLayer(
                num_heads=num_heads,
                num_channels=num_channels,
                num_qk_channels=num_qk_channels,
                num_v_channels=num_v_channels,
                max_heads_parallel=max_heads_parallel,
                causal_attention=causal_attention,
                widening_factor=widening_factor,
                dropout=dropout,
                qkv_bias=qkv_bias,
                out_bias=out_bias,
                mlp_bias=mlp_bias,
            )
            for _ in range(num_layers)
        ]

        if activation_checkpointing:
            layers = [checkpoint_wrapper(layer, offload_to_cpu=activation_offloading) for layer in layers]

        super().__init__(*layers)


class MLP(Sequential):
    def __init__(self, num_channels: int, widening_factor: int, bias: bool = True):
        super().__init__(
            nn.LayerNorm(num_channels),
            nn.Linear(num_channels, widening_factor * num_channels, bias=bias),
            nn.GELU(),
            nn.Linear(widening_factor * num_channels, num_channels, bias=bias),
        )


class PerceiverEncoder(nn.Module):
    def __init__(
        self,
        input_adapter: InputAdapter,
        num_latents: int,
        num_latent_channels: int,
        num_cross_attention_heads: int = 4,
        num_cross_attention_qk_channels: Optional[int] = None,
        num_cross_attention_v_channels: Optional[int] = None,
        num_cross_attention_layers: int = 1,
        first_cross_attention_layer_shared: bool = False,
        cross_attention_widening_factor: int = 1,
        num_self_attention_heads: int = 4,
        num_self_attention_qk_channels: Optional[int] = None,
        num_self_attention_v_channels: Optional[int] = None,
        num_self_attention_layers_per_block: int = 6,
        num_self_attention_blocks: int = 1,
        first_self_attention_block_shared: bool = True,
        self_attention_widening_factor: int = 1,
        dropout: float = 0.0,
        init_scale: float = 0.02,
        activation_checkpointing: bool = False,
        activation_offloading: bool = False,
    ):
        """Generic Perceiver IO encoder.

        :param input_adapter: Transforms and position-encodes task-specific input to generic encoder input
            of shape (B, M, C) where B is the batch size, M the input sequence length and C the number of
            key/value input channels. C is determined by the `num_input_channels` property of the
            `input_adapter`.
        :param num_latents: Number of latent variables (N).
        :param num_latent_channels: Number of latent channels (D).
        :param num_cross_attention_heads: Number of cross-attention heads.
        :param num_cross_attention_qk_channels: Number of query and key channels for cross-attention
            (see `MultiHeadAttention.num_qk_channels` for details).
        :param num_cross_attention_v_channels: Number of value channels for cross-attention
            (see `MultiHeadAttention.num_v_channels` for details).
        :param num_cross_attention_layers: Number of cross-attention layers (alternating with self-attention blocks).
        :param first_cross_attention_layer_shared: Whether the first cross-attention layer should share its weights
            with subsequent cross-attention layers (if any).
        :param num_self_attention_heads: Number of self-attention heads.
        :param num_self_attention_qk_channels: Number of query and key channels for self-attention
            (see `MultiHeadAttention.num_qk_channels` for details).
        :param num_self_attention_v_channels: Number of value channels for self-attention
            (see `MultiHeadAttention.num_v_channels` for details).
        :param num_self_attention_layers_per_block: Number of self-attention layers per self-attention block.
        :param num_self_attention_blocks: Number of self-attention blocks, with weights shared between corresponding
            self-attention layers.
        :param first_self_attention_block_shared: Whether the first self-attention block should share its weights
            with subsequent self-attention blocks (if any).
        :param dropout: Dropout probability for self- and cross-attention layers.
        :param init_scale: Standard deviation for random normal initialization of parameters.
        :param activation_checkpointing: If True, implements an activation checkpoint for each self-attention
            layer and each cross-attention layer.
        :param activation_offloading: If True, offloads checkpointed activations to CPU.
        """
        super().__init__()

        self.latent_provider = TrainableQueryProvider(num_latents, num_latent_channels, init_scale=init_scale)
        self.input_adapter = input_adapter

        if num_cross_attention_layers <= 0:
            raise ValueError("num_cross_attention_layers must be > 0")

        if num_self_attention_blocks <= 0:
            raise ValueError("num_self_attention_blocks must be > 0")

        if num_cross_attention_layers > num_self_attention_blocks:
            raise ValueError("num_cross_attention_layers must be <= num_self_attention_blocks")

        self.num_cross_attention_layers = num_cross_attention_layers
        self.num_self_attention_blocks = num_self_attention_blocks

        self.first_cross_attention_layer_shared = first_cross_attention_layer_shared
        self.first_self_attention_block_shared = first_self_attention_block_shared

        def cross_attn():
            layer = CrossAttentionLayer(
                num_heads=num_cross_attention_heads,
                num_q_input_channels=num_latent_channels,
                num_kv_input_channels=input_adapter.num_input_channels,
                num_qk_channels=num_cross_attention_qk_channels,
                num_v_channels=num_cross_attention_v_channels,
                widening_factor=cross_attention_widening_factor,
                dropout=dropout,
            )
            return (
                checkpoint_wrapper(layer, offload_to_cpu=activation_offloading) if activation_checkpointing else layer
            )

        def self_attn():
            return SelfAttentionBlock(
                num_layers=num_self_attention_layers_per_block,
                num_heads=num_self_attention_heads,
                num_channels=num_latent_channels,
                num_qk_channels=num_self_attention_qk_channels,
                num_v_channels=num_self_attention_v_channels,
                widening_factor=self_attention_widening_factor,
                dropout=dropout,
                activation_checkpointing=activation_checkpointing,
                activation_offloading=activation_offloading,
            )

        self.cross_attn_1 = cross_attn()
        self.self_attn_1 = self_attn()

        if self.extra_cross_attention_layer:
            self.cross_attn_n = cross_attn()

        if self.extra_self_attention_block:
            self.self_attn_n = self_attn()

        self._init_parameters(init_scale)

    def _init_parameters(self, init_scale: float):
        with torch.no_grad():
            init_parameters(self, init_scale)

    @property
    def extra_cross_attention_layer(self):
        return self.num_cross_attention_layers > 1 and not self.first_cross_attention_layer_shared

    @property
    def extra_self_attention_block(self):
        return self.num_self_attention_blocks > 1 and not self.first_self_attention_block_shared

    def forward(self, x, pad_mask=None, return_adapted_input=False):
        b, *_ = x.shape

        x_adapted = self.input_adapter(x)
        x_latent = self.latent_provider()

        x_latent = self.cross_attn_1(x_latent, x_adapted, pad_mask=pad_mask)
        x_latent = self.self_attn_1(x_latent)

        cross_attn_n = self.cross_attn_n if self.extra_cross_attention_layer else self.cross_attn_1
        self_attn_n = self.self_attn_n if self.extra_self_attention_block else self.self_attn_1

        for i in range(1, self.num_self_attention_blocks):
            if i < self.num_cross_attention_layers:
                x_latent = cross_attn_n(x_latent, x_adapted, pad_mask=pad_mask)
            x_latent = self_attn_n(x_latent)

        if return_adapted_input:
            return x_latent, x_adapted
        else:
            return x_latent


class PerceiverDecoder(nn.Module):
    def __init__(
        self,
        output_adapter: OutputAdapter,
        output_query_provider: QueryProvider,
        num_latent_channels: int,
        num_cross_attention_heads: int = 4,
        num_cross_attention_qk_channels: Optional[int] = None,
        num_cross_attention_v_channels: Optional[int] = None,
        cross_attention_widening_factor: int = 1,
        cross_attention_residual: bool = True,
        dropout: float = 0.0,
        init_scale: float = 0.02,
        activation_checkpointing: bool = False,
        activation_offloading: bool = False,
    ):
        """Generic Perceiver IO decoder.

        :param output_adapter: Transforms generic decoder cross-attention output of shape (B, O, F) to task-specific
            output. B is the batch size, O the output sequence length and F the number of cross-attention output
            channels.
        :param output_query_provider: Provides the decoder's output query. Abstracts over output query details e.g.
            can be a learned query, a deterministic function of the model's input, etc. Configured by `PerceiverIO`
            subclasses.
        :param num_latent_channels: Number of latent channels of the Perceiver IO encoder output.
        :param num_cross_attention_heads: Number of cross-attention heads.
        :param num_cross_attention_qk_channels: Number of query and key channels for cross-attention
            (see `MultiHeadAttention.num_qk_channels` for details).
        :param num_cross_attention_v_channels: Number of value channels for cross-attention
            (see `MultiHeadAttention.num_v_channels` for details).
        :param dropout: Dropout probability for cross-attention layer.
        :param init_scale: Standard deviation for random normal initialization of parameters.
        :param activation_checkpointing: If True, implements an activation checkpoint for the decoder's
            cross-attention layer.
        :param activation_offloading: If True, offloads checkpointed activations to CPU.
        """
        super().__init__()

        self.output_query_provider = output_query_provider
        self.output_adapter = output_adapter

        cross_attn = CrossAttentionLayer(
            num_heads=num_cross_attention_heads,
            num_q_input_channels=output_query_provider.num_query_channels,
            num_kv_input_channels=num_latent_channels,
            num_qk_channels=num_cross_attention_qk_channels,
            num_v_channels=num_cross_attention_v_channels,
            widening_factor=cross_attention_widening_factor,
            attention_residual=cross_attention_residual,
            dropout=dropout,
        )

        if activation_checkpointing:
            cross_attn = checkpoint_wrapper(cross_attn, offload_to_cpu=activation_offloading)

        self.cross_attn = cross_attn
        self._init_parameters(init_scale)

    def _init_parameters(self, init_scale: float):
        with torch.no_grad():
            init_parameters(self, init_scale)

    def forward(self, x_latent, x_adapted=None, **kwargs):
        output_query = self.output_query_provider(x_adapted)
        output = self.cross_attn(output_query, x_latent)
        return self.output_adapter(output, **kwargs)


class PerceiverIO(Sequential):
    def __init__(self, encoder: PerceiverEncoder, decoder: PerceiverDecoder):
        super().__init__(encoder, decoder)

    @property
    def encoder(self):
        return self[0]

    @property
    def decoder(self):
        return self[1]


class PerceiverAR(nn.Module):
    def __init__(
        self,
        input_adapter: RotarySupport,
        num_heads: int = 8,
        max_heads_parallel: Optional[int] = None,
        num_self_attention_layers: int = 6,
        self_attention_widening_factor: int = 4,
        cross_attention_widening_factor: int = 4,
        cross_attention_dropout: float = 0.5,
        post_attention_dropout: float = 0.0,
        activation_checkpointing: bool = False,
        activation_offloading: bool = False,
    ):
        """Implementation of Perceiver AR (https://arxiv.org/abs/2202.07765).

        :param input_adapter: Transforms an input sequence to generic Perceiver AR input. An input adapter may choose
            to add (absolute) position information to transformed inputs while `PerceiverAR` additionally computes a
            rotary position embedding (i.e. relative position information) for queries and keys. To support the
            computation of rotary position embeddings, concrete input adapters need to mixin `RotarySupport`.
        :param num_heads: Number of cross- and self-attention heads.
        :param max_heads_parallel: Maximum number of cross-attention heads to be processed in parallel.
            Default is `num_heads`.
        :param num_self_attention_layers: Number of self-attention layers.
        :param cross_attention_dropout: Probability of dropping positions in the prefix sequence.
        :param post_attention_dropout: Probability of dropping cross- and self-attention scores (same as `dropout`
            in Perceiver IO encoder and decoder).
        :param activation_checkpointing: If True, implements an activation checkpoint for each self-attention
            layer and cross-attention layer.
        :param activation_offloading: If True, offloads checkpointed activations to CPU.
        """
        super().__init__()

        def cross_attn():
            layer = CrossAttentionLayer(
                num_heads=num_heads,
                num_q_input_channels=input_adapter.num_input_channels,
                num_kv_input_channels=input_adapter.num_input_channels,
                max_heads_parallel=max_heads_parallel,
                causal_attention=True,
                widening_factor=cross_attention_widening_factor,
                dropout=post_attention_dropout,
                qkv_bias=False,
                out_bias=True,
                mlp_bias=False,
            )
            return (
                checkpoint_wrapper(layer, offload_to_cpu=activation_offloading) if activation_checkpointing else layer
            )

        def self_attn():
            return SelfAttentionBlock(
                num_layers=num_self_attention_layers,
                num_heads=num_heads,
                num_channels=input_adapter.num_input_channels,
                causal_attention=True,
                widening_factor=self_attention_widening_factor,
                dropout=post_attention_dropout,
                activation_checkpointing=activation_checkpointing,
                activation_offloading=activation_offloading,
                qkv_bias=False,
                out_bias=False,
                mlp_bias=False,
            )

        self.input_adapter = input_adapter

        self.cross_attention_dropout = cross_attention_dropout
        self.cross_attention = cross_attn()
        self.self_attention = self_attn()

    def forward(self, x, prefix_len, pad_mask=None):
        b, n = x.shape

        if not 0 <= prefix_len < n:
            raise ValueError(f"prefix_len ({prefix_len}) out of valid range [0..{n})")

        if pad_mask is None:
            shift = None
        else:
            # caller must ensure that x is left-padded
            shift = pad_mask.sum(dim=1, keepdim=True)

        # freq_pos_enc shape is (b, n, f)
        x, frq_pos_enc = self.input_adapter(x, abs_pos=positions(*x.shape, shift=shift, device=x.device))

        x_latent = x[:, prefix_len:]
        x_prefix = x[:, :prefix_len]

        frq_pos_enc_latent = frq_pos_enc[:, prefix_len:]
        frq_pos_enc_prefix = frq_pos_enc[:, :prefix_len]

        if pad_mask is not None:
            pad_mask_latent = pad_mask[:, prefix_len:]
            pad_mask_prefix = pad_mask[:, :prefix_len]

        if self.training and prefix_len > 0 and self.cross_attention_dropout > 0.0:
            # Modified from https://github.com/lucidrains/perceiver-ar-pytorch
            rand = torch.rand(b, prefix_len, device=x.device)
            # number of positions in prefix sequence to keep
            keep = prefix_len - int(prefix_len * self.cross_attention_dropout)
            # indices of positions in prefix sequence to keep
            keep_indices = rand.topk(keep, dim=-1).indices
            # mask of positions in prefix sequence to keep
            keep_mask = torch.zeros_like(rand, dtype=torch.bool).scatter_(dim=1, index=keep_indices, value=1)

            # drop positions in prefix sequence according to prefix_dropout
            x_prefix = rearrange(x_prefix[keep_mask], "(b n) c -> b n c", b=b)
            # drop positions in prefix frequency position encoding
            frq_pos_enc_prefix = rearrange(frq_pos_enc_prefix[keep_mask], "(b n) f -> b n f", b=b)

            if pad_mask is not None:
                # drop positions in prefix pad mask
                pad_mask_prefix = rearrange(pad_mask_prefix[keep_mask], "(b n) -> b n", b=b)

        frq_pos_enc_q = frq_pos_enc_latent
        frq_pos_enc_k = torch.cat([frq_pos_enc_prefix, frq_pos_enc_latent], dim=1)

        if pad_mask is not None:
            pad_mask = torch.cat([pad_mask_prefix, pad_mask_latent], dim=1)

        x_latent = self.cross_attention(
            x_latent,
            x_kv_prefix=x_prefix,
            pad_mask=pad_mask,
            rot_pos_emb_q=RotaryPositionEmbedding(frq_pos_enc_q, right_align=True),
            rot_pos_emb_k=RotaryPositionEmbedding(frq_pos_enc_k, right_align=True),
        )

        x_latent = self.self_attention(
            x_latent,
            rot_pos_emb=RotaryPositionEmbedding(frq_pos_enc_latent, right_align=True),
        )

        return x_latent



import html
import os
from dataclasses import dataclass
from typing import Any, List, Optional

import torch
import torch.nn as nn
import transformers
from einops import rearrange
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.utilities import rank_zero_only

from perceiver.model.core import (
    DecoderConfig,
    LitPerceiverIO,
    OutputAdapter,
    PerceiverDecoder,
    PerceiverIO,
    PerceiverIOConfig,
    TrainableQueryProvider,
)

from perceiver.model.core.convert import copy_cross_attention_layer_params, copy_param
from perceiver.model.core.utils import is_checkpoint
from perceiver.model.text.common import TextEncoder, TextEncoderConfig, TiedTextOutputAdapter


@dataclass
class TextDecoderConfig(DecoderConfig):
    num_output_query_channels: Optional[int] = None
    vocab_size: int = 10003
    max_seq_len: int = 512


class TextOutputAdapter(OutputAdapter):
    def __init__(self, vocab_size: int, num_output_query_channels: int):
        super().__init__()
        self.linear = nn.Linear(num_output_query_channels, vocab_size)

    def forward(self, x):
        return self.linear(x).squeeze(dim=1)


class MaskedLanguageModel(PerceiverIO):
    def __init__(self, config: PerceiverIOConfig[TextEncoderConfig, TextDecoderConfig]):
        encoder = TextEncoder(
            config.encoder,
            num_latents=config.num_latents,
            num_latent_channels=config.num_latent_channels,
            activation_checkpointing=config.activation_checkpointing,
            activation_offloading=config.activation_offloading,
        )
        if config.decoder.num_output_query_channels is None:
            output_query_provider = TrainableQueryProvider(
                num_queries=config.decoder.max_seq_len,
                num_query_channels=config.encoder.num_input_channels,
                init_scale=config.decoder.init_scale,
            )
            output_adapter = TiedTextOutputAdapter(
                vocab_size=config.decoder.vocab_size,
            )
        else:
            output_query_provider = TrainableQueryProvider(
                num_queries=config.decoder.max_seq_len,
                num_query_channels=config.decoder.num_output_query_channels,
                init_scale=config.decoder.init_scale,
            )
            output_adapter = TextOutputAdapter(
                vocab_size=config.decoder.vocab_size,
                num_output_query_channels=config.decoder.num_output_query_channels,
            )
        decoder = TextDecoder(
            output_adapter=output_adapter,
            output_query_provider=output_query_provider,
            num_latent_channels=config.num_latent_channels,
            activation_checkpointing=config.activation_checkpointing,
            activation_offloading=config.activation_offloading,
            **config.decoder.base_kwargs()
        )
        super().__init__(encoder, decoder)

        if config.params is None or is_checkpoint(config.params):
            pass
        elif os.path.isfile(config.params):
            self.load_state_dict(torch.load(config.params))
        else:
            # import model params from Hugging Face Perceiver
            src = transformers.PerceiverForMaskedLM.from_pretrained(config.params)
            self.encoder.copy_params(src.perceiver)
            self.decoder.copy_params(src)

    def forward(self, x_masked, pad_mask=None):
        _, n = x_masked.shape

        x_latent = self.encoder(x_masked, pad_mask)
        if isinstance(self.decoder.output_adapter, TiedTextOutputAdapter):
            x_logits = self.decoder(x_latent, txt_embedding=self.encoder.input_adapter.txt_embedding)
        else:
            x_logits = self.decoder(x_latent)

        return x_logits[:, :n, :]


class TextDecoder(PerceiverDecoder):
    def copy_params(self, src: transformers.PerceiverForMaskedLM):
        copy_cross_attention_layer_params(
            src.perceiver.decoder.decoding_cross_attention, self.cross_attn, query_residual=False
        )
        # Copy output query provider parameters
        copy_param(
            src.perceiver.decoder.output_position_encodings.position_embeddings, self.output_query_provider._query
        )
        # Copy output adapter parameters
        copy_param(src.embedding_decoder.bias, self.output_adapter.bias)


class LitMaskedLanguageModel(LitPerceiverIO):
    def __init__(
        self,
        encoder: TextEncoderConfig,
        decoder: TextDecoderConfig,
        *args: Any,
        num_predictions: int = 3,
        masked_samples: Optional[List[str]] = None,
        **kwargs: Any
    ):
        super().__init__(encoder, decoder, *args, **kwargs)
        self.model = MaskedLanguageModel(
            PerceiverIOConfig(
                encoder=encoder,
                decoder=decoder,
                num_latents=self.hparams.num_latents,
                num_latent_channels=self.hparams.num_latent_channels,
                activation_checkpointing=self.hparams.activation_checkpointing,
                activation_offloading=self.hparams.activation_offloading,
                params=self.hparams.params,
            )
        )
        self.loss = nn.CrossEntropyLoss()

        if self.hparams.params is not None and is_checkpoint(self.hparams.params):
            lit_model = LitMaskedLanguageModel.load_from_checkpoint(self.hparams.params, params=None)
            self.model.load_state_dict(lit_model.model.state_dict())

    def setup(self, stage: Optional[str] = None):
        self.filler = MaskedSampleFiller(preprocessor=self.trainer.datamodule.text_preprocessor())

    def forward(self, x, pad_mask):
        return self.model(x, pad_mask)

    def step(self, batch):
        labels, x, pad_mask = batch
        logits = self(x, pad_mask)
        logits = rearrange(logits, "b n c -> (b n) c")
        labels = rearrange(labels, "b n -> (b n)")
        return self.loss(logits, labels)

    def training_step(self, batch, batch_idx):
        loss = self.step(batch)
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        loss = self.step(batch)
        self.log("val_loss", loss, prog_bar=True, sync_dist=True)

    def test_step(self, batch, batch_idx):
        loss = self.step(batch)
        self.log("test_loss", loss, sync_dist=True)

    @rank_zero_only
    def on_validation_epoch_end(self) -> None:
        if self.hparams.masked_samples:
            masked_samples, filled_samples = self.filler.fill(
                model=self,
                masked_samples=self.hparams.masked_samples,
                num_predictions=self.hparams.num_predictions,
                device=self.device,
            )

            if isinstance(self.logger, TensorBoardLogger):
                rendered_samples = "\n\n".join(
                    ["  \n".join([html.escape(s)] + ps) for s, ps in zip(masked_samples, filled_samples)]
                )
                self.logger.experiment.add_text("sample predictions", rendered_samples, self.trainer.global_step)
            else:
                # support other loggers here ...
                ...


class MaskedSampleFiller:
    def __init__(self, preprocessor):
        self.preprocessor = preprocessor

    def fill(self, model, masked_samples, num_predictions, device="cpu"):
        masked_samples = [ms.replace("<mask>", self.preprocessor.tokenizer.mask_token) for ms in masked_samples]

        xs, ms = self.preprocessor.preprocess_batch(masked_samples)
        xs = xs.to(device)
        ms = ms.to(device)

        with torch.no_grad():
            x_logits = model(xs, ms)

        pred_mask = xs == self.preprocessor.tokenizer.mask_token_id
        pred_ids = torch.topk(x_logits[pred_mask, :], k=num_predictions, dim=1).indices

        results = []

        for i in range(num_predictions):
            xs[pred_mask] = pred_ids[:, i]
            results.append(self.preprocessor.tokenizer.batch_decode(xs, skip_special_tokens=True))

        return masked_samples, list(map(list, zip(*results)))  # transpose results (a list of lists)


def convert_config(config: transformers.PerceiverConfig) -> PerceiverIOConfig[TextEncoderConfig, TextDecoderConfig]:
    assert config.hidden_act == "gelu"
    assert config.tie_word_embeddings

    encoder_config = TextEncoderConfig(
        vocab_size=config.vocab_size,
        max_seq_len=config.max_position_embeddings,
        num_input_channels=config.d_model,
        num_cross_attention_qk_channels=config.qk_channels,
        num_cross_attention_v_channels=config.v_channels,
        num_cross_attention_heads=config.num_cross_attention_heads,
        num_self_attention_qk_channels=config.qk_channels,
        num_self_attention_v_channels=config.v_channels,
        num_self_attention_heads=config.num_self_attention_heads,
        num_self_attention_layers_per_block=config.num_self_attends_per_block,
        num_self_attention_blocks=config.num_blocks,
        cross_attention_widening_factor=config.cross_attention_widening_factor,
        self_attention_widening_factor=config.self_attention_widening_factor,
        dropout=config.attention_probs_dropout_prob,
        init_scale=config.initializer_range,
    )
    decoder_config = TextDecoderConfig(
        vocab_size=config.vocab_size,
        max_seq_len=config.max_position_embeddings,
        num_cross_attention_qk_channels=config.qk_channels,
        num_cross_attention_v_channels=config.d_model,
        num_cross_attention_heads=config.num_cross_attention_heads,
        cross_attention_widening_factor=config.cross_attention_widening_factor,
        cross_attention_residual=False,
        dropout=config.attention_probs_dropout_prob,
        init_scale=config.initializer_range,
    )
    return PerceiverIOConfig(
        encoder_config,
        decoder_config,
        num_latents=config.num_latents,
        num_latent_channels=config.d_latents,
        params=config.name_or_path,
    )



import os.path
from pathlib import Path
from typing import List, Tuple

import cv2
import numpy as np


def read_video_frames(video_path: Path) -> List[np.ndarray]:
    if not video_path.exists():
        raise ValueError(f"Invalid video path supplied. Path '{video_path}' does not exist.")

    frames = []
    cap = cv2.VideoCapture(str(video_path))
    try:
        while cap.isOpened():
            ret, frame = cap.read()

            if not ret:
                break

            frames.append(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
    finally:
        if cap is not None:
            cap.release()

    return frames


def read_video_frame_pairs(video_path: Path) -> List[Tuple[np.ndarray, np.ndarray]]:
    frames = read_video_frames(video_path)

    return list(zip(frames, frames[1:]))


def write_video(video_path: Path, frames: List[np.ndarray], fps: int) -> None:
    _, ext = os.path.splitext(video_path)
    if ext != ".mp4":
        raise ValueError("Invalid video path supplied. Only files of type 'mp4' are supported.")

    frame_shape = frames[0].shape

    video = cv2.VideoWriter(str(video_path), cv2.VideoWriter_fourcc(*"mp4v"), fps, (frame_shape[1], frame_shape[0]))
    for frame in frames:
        video.write(cv2.cvtColor(frame, cv2.COLOR_RGB2BGR))

    video.release()



import torch


class ImagePreprocessor:
    def __init__(self, transform):
        self.transform = transform

    def preprocess(self, img):
        return self.transform(img)

    def preprocess_batch(self, img_batch):
        return torch.stack([self.preprocess(img) for img in img_batch])


def lift_transform(transform):
    def apply(examples):
        examples["image"] = [transform(image) for image in examples["image"]]
        return examples

    return apply


def channels_to_last(img: torch.Tensor):
    return img.permute(1, 2, 0).contiguous()



import os
from dataclasses import dataclass
from typing import Optional

import torch
import torch.nn as nn
import transformers

from perceiver.model.core import EncoderConfig, InputAdapter, OutputAdapter, PerceiverEncoder
from perceiver.model.core.convert import (
    copy_cross_attention_layer_params,
    copy_latent_provider_params,
    copy_params,
    copy_self_attention_block_params,
)
from perceiver.model.core.position import positions
from perceiver.model.core.utils import freeze, is_checkpoint


@dataclass
class TextEncoderConfig(EncoderConfig):
    vocab_size: int = 10003
    max_seq_len: int = 256
    num_input_channels: int = 64
    params: Optional[str] = None


class TextInputAdapter(InputAdapter):
    def __init__(self, vocab_size: int, max_seq_len: int, num_input_channels: int):
        super().__init__(num_input_channels)
        self.txt_embedding = nn.Embedding(vocab_size, num_input_channels)
        self.pos_embedding = nn.Embedding(max_seq_len, num_input_channels)

    @property
    def vocab_size(self):
        return self.txt_embedding.num_embeddings

    @property
    def max_seq_len(self):
        return self.pos_embedding.num_embeddings

    def forward(self, x, abs_pos=None):
        if abs_pos is None:
            abs_pos = positions(*x.shape, device=x.device)
        return self.txt_embedding(x) + self.pos_embedding(abs_pos)


class TiedTextOutputAdapter(OutputAdapter):
    def __init__(self, vocab_size: int):
        super().__init__()
        self.bias = nn.Parameter(torch.zeros(vocab_size))

    def forward(self, x, txt_embedding: nn.Embedding):
        return torch.matmul(x, txt_embedding.weight.T) + self.bias


class TextEncoder(PerceiverEncoder):
    def __init__(
        self,
        config: TextEncoderConfig,
        num_latents: int,
        num_latent_channels: int,
        activation_checkpointing: bool,
        activation_offloading: bool,
    ):
        input_adapter = TextInputAdapter(
            vocab_size=config.vocab_size,
            max_seq_len=config.max_seq_len,
            num_input_channels=config.num_input_channels,
        )
        super().__init__(
            input_adapter=input_adapter,
            num_latents=num_latents,
            num_latent_channels=num_latent_channels,
            activation_checkpointing=activation_checkpointing,
            activation_offloading=activation_offloading,
            **config.base_kwargs()
        )

        if config.params is None or is_checkpoint(config.params):
            pass
        elif os.path.isfile(config.params):
            self.load_state_dict(torch.load(config.params))
        else:
            # import encoder params from Hugging Face Perceiver
            self.copy_params(transformers.PerceiverForMaskedLM.from_pretrained(config.params))

        if config.freeze:
            freeze(self)

    def copy_params(self, src: transformers.PerceiverModel):
        copy_cross_attention_layer_params(src.encoder.cross_attention, self.cross_attn_1, query_residual=True)
        copy_self_attention_block_params(src.encoder.self_attends, self.self_attn_1)
        copy_latent_provider_params(src, self)
        # Copy input adapter parameters
        copy_params(src.input_preprocessor.embeddings, self.input_adapter.txt_embedding)
        copy_params(src.input_preprocessor.position_embeddings, self.input_adapter.pos_embedding)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['perceiver',
 'perceiver.data',
 'perceiver.data.text',
 'perceiver.data.vision',
 'perceiver.model',
 'perceiver.model.core',
 'perceiver.model.text',
 'perceiver.model.vision',
 'perceiver.scripts',
 'perceiver.scripts.text',
 'perceiver.scripts.vision']

package_data = \
{'': ['*']}

install_requires = \
['cchardet>=2.1,<3.0',
 'einops>=0.4,<0.5',
 'fairscale>=0.4,<0.5',
 'fsspec[s3]',
 'jsonargparse[signatures]>=4.12,<5.0',
 'pytorch-lightning>=1.7,<2.0',
 'tensorboard>=2.11,<3.0',
 'torch-optimizer>=0.3,<0.4',
 'torch>=1.13,<2.0',
 'torchmetrics>=0.9,<0.10']

extras_require = \
{'text': ['datasets>=2.4,<3.0',
          'tokenizers>=0.12,<0.13',
          'transformers>=4.21,<5.0'],
 'vision': ['datasets>=2.4,<3.0',
            'torchvision>=0.14,<0.15',
            'opencv-python>=4.6.0.66,<5.0.0.0']}

setup_kwargs = {
    'name': 'perceiver-io',
    'version': '0.8.1',
    'description': 'Perceiver IO',
    'long_description': '# Perceiver, Perceiver IO and Perceiver AR\n\nThis repository is a PyTorch and PyTorch Lightning implementation of\n\n<table>\n  <tr>\n    <td>\n       <b>Perceiver</b>: General Perception with Iterative Attention\n       (<a href="https://arxiv.org/abs/2103.03206">paper</a>,\n        <a href="https://www.youtube.com/watch?v=P_xeshTnPZg">video</a>)\n    </td>\n    <td><img src="docs/images/small-perceiver.png" alt="Perceiver"/></td>\n  </tr>\n  <tr>\n    <td>\n      <b>Perceiver IO</b>: A General Architecture for Structured Inputs & Outputs\n      (<a href="https://arxiv.org/abs/2107.14795">paper</a>,\n       <a href="https://www.deepmind.com/blog/building-architectures-that-can-handle-the-worlds-data">blog post</a>)\n    </td>\n    <td><img src="docs/images/small-perceiver-io.png" alt="Perceiver IO"/></td>\n  </tr>\n  <tr>\n    <td>\n      General-purpose, long-context autoregressive modeling with <b>Perceiver AR</b>\n      (<a href="https://arxiv.org/abs/2202.07765">paper</a>,\n       <a href="https://www.deepmind.com/blog/perceiver-ar-general-purpose-long-context-autoregressive-generation">blog post</a>)\n    </td>\n    <td><img src="docs/images/small-perceiver-ar.png" alt="Perceiver AR"/></td>\n  </tr>\n</table>\n\nAll model classes are written in plain PyTorch and can be wrapped into [PyTorch Lightning](https://pytorch-lightning.readthedocs.io/en/stable/)\nmodules for training at scale. The command line interface is implemented with the [Lightning CLI](https://pytorch-lightning.readthedocs.io/en/stable/cli/lightning_cli.html).\n[Pretrained weights](docs/pretrained-models.md) can be imported for [official models](docs/pretrained-models.md#official-models)\nfrom the 🤗 Hub, [training checkpoints](docs/pretrained-models.md#training-checkpoints) from [training examples](docs/training-examples.md)\nare available for download too. Datasets used in the training examples are 🤗 [datasets](https://huggingface.co/docs/datasets)\nwrapped into PyTorch Lightning [data modules](perceiver/data). For NLP tasks, this library supports all 🤗\n[fast tokenizers](https://huggingface.co/docs/transformers/fast_tokenizers) and the 🤗 Perceiver UTF-8 bytes tokenizer.\n\n## Installation\n\n### Via pip\n\n```shell\npip install perceiver-io[text,vision]\n```\n\n### From sources\n\nInstallation from sources requires a [Miniconda](https://docs.conda.io/en/latest/miniconda.html) and a\n[Poetry](https://python-poetry.org/docs/#installation) (1.2.0 or higher) installation.\n\nCreate and activate the `perceiver-io` conda environment:\n\n```shell\nconda env create -f environment.yml\nconda activate perceiver-io\n```\n\nInstall main and test dependencies, including all extras:\n\n```shell\n# Without dependencies required for examples\npoetry install --all-extras\n```\n\nIf you want to run the [examples](examples) locally, additionally use `--with examples`:\n\n```shell\npoetry install --all-extras --with examples\n```\n\n### Docker image\n\n```shell\ndocker pull ghcr.io/krasserm/perceiver-io:latest\n```\n\nSee [Docker image](docs/docker-image.md) for details.\n\n## Documentation\n\n- [Getting started](docs/getting-started.md)\n- [Model construction](docs/model-construction.md)\n- [Pretrained models](docs/pretrained-models.md)\n- [Training examples](docs/training-examples.md)\n- [Inference examples](examples/inference.ipynb) [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/krasserm/perceiver-io/blob/0.8.0/examples/inference.ipynb)\n- [Building blocks](docs/building-blocks.md)\n\n## Articles\n\nArticles referencing this repository:\n\n- [Training compute-optimal Perceiver AR language models](https://krasserm.github.io/2023/01/23/scaling-perceiver-ar/)\n- [A gentle introduction to Rotary Position Embedding](https://krasserm.github.io/2022/12/13/rotary-position-embedding/)\n\n## Other implementations\n\n- [Perceiver](https://paperswithcode.com/paper/perceiver-general-perception-with-iterative#code)\n- [Perceiver IO](https://paperswithcode.com/paper/perceiver-io-a-general-architecture-for#code)\n- [Perceiver AR](https://paperswithcode.com/paper/general-purpose-long-context-autoregressive#code)\n',
    'author': 'Martin Krasser',
    'author_email': 'krasserm@googlemail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/krasserm/perceiver-io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import math

from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR


class CosineWithWarmupLR(LambdaLR):
    def __init__(
        self,
        optimizer: Optimizer,
        training_steps: int = 0,
        warmup_steps: int = 0,
        num_cycles: float = 0.5,
        min_fraction: float = 0.0,
        last_epoch: int = -1,
    ):
        # Can be updated after instantiation
        self.training_steps = training_steps

        def lr_lambda(current_step):
            if current_step < warmup_steps:
                return float(current_step) / float(max(1, warmup_steps))
            progress = float(current_step - warmup_steps) / float(max(1, self.training_steps - warmup_steps))
            return min_fraction + max(
                0.0, 0.5 * (1.0 - min_fraction) * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress))
            )

        super().__init__(optimizer, lr_lambda, last_epoch=last_epoch)


class ConstantWithWarmupLR(LambdaLR):
    def __init__(self, optimizer: Optimizer, warmup_steps: int = 0, last_epoch: int = -1):
        def lr_lambda(current_step: int):
            if current_step < warmup_steps:
                return float(current_step) / float(max(1.0, warmup_steps))
            return 1.0

        super().__init__(optimizer, lr_lambda, last_epoch=last_epoch)
